# regplug.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

# Plugin registration. 
# plugin() is called by the PLUGIN command. It searchs and loads
# the required plugin code and adds them to the command tables in parse.py.

import os
import sys
import importlib
import hashlib
import json
import mma.MMA.debug

from mma.MMA.paths import plugPaths
from mma.MMA.common import *
import mma.MMA.parse
from mma.MMA.alloc import trkClasses
import mma.MMA.appdirs   # not mine, but it works!

# In python3 raw_input() has been renamed input()
if sys.version_info[0] == 3:
    raw_input = input

# Set the name of the entry point module. Each plugin must
# have this module (with a .py ending) in its directory.
# Filename is case-insensitive
entry = "plugin"

# This is prepended to the module name to make it unique
# as a command name. The module "foo.py" becomes the command
# "@foo"
prefix = "@"

# these are for the macro funcs and for debug.
plugList = []      # all loaded plugins
simplePlugs = []   # simple funcs registered
trackPlugs = []    # track funcs registered
dataPlugs = []     # data plugs registered

# A list of entry points for the command line help
# function. Each plugin should have a function printUsage().
# If this function is found, we add it. The -I <plugin>
# command line switch will call/display it.
plugHelp = {}

# Security permission file of accepted plugins.
# The path/file is set in getPermission() function if needed.
permFile = None

# This can be set to TRUE from the cmd line with -II
# If set the security file is ignored.
secOverRide=False

# For security we disable various load paths. These can
# disabled via the Plugin Disable=... command. They can
# not be enabled!
tryLocal = True     # from the same dir as the current file
tryDot = True       # from the users current dir
tryPlugDir = True   # from the plugin directory
plugsOff = False    # disable plug loading completely

def findPlugin(targ):
    """ Search for the plugin. 

        p - the plugin name (eg. hello)

        We search, in order, the following paths for the file hello/plugin.py
    
           - the users current directory
           - the directory in which the current file being processed lives
           - the plugin directories.

           The search is case-insensitive.

        Returns - complete path to the directory containing "hello/plugin.py",
                  and the module name (hello.plugin).
    """

    plugEntry = "%s.%s" % (entry, 'py')
    
    def matchName(d):
        """ Find a directory entry in 'd' matching the plugin name """

        if os.access(d, os.R_OK):  # skip non-exist dirs and ones we can't access
            if [b.upper() for b in os.listdir(d)].count(targ) > 1:
                warning("Plugin: there are duplicate entries of '%s' in '%s'. This "
                   "is most likly due to the same name with different cases. You "
                   "should check and delete/rename one!" % (targ, d))

            for a in os.listdir(d):
                if a.lower() == targ.lower():
                    for b in os.listdir(os.path.join(d, a)):
                        if b.lower() == plugEntry.lower():
                            # Create a module name based on the case of the files found
                            mName = "%s.%s" % (a, b[0:-3])
                            return os.path.join(a, b), mName

        return None, None

    # 1. Current dir. If found convert to complete path
    if tryDot:
        for dir in ('.', 'plugins'):
            mdir, modName = matchName(dir)
            if mdir:
                return os.path.realpath(dir), mdir, modName

    # 2. Check for the plugin in the same directory as the current
    #    file being processed ... the current file could be a library file.
    if tryLocal:
        if gbl.inpath:
            n = gbl.inpath.fname
            path = os.path.dirname(gbl.inpath.fname)
            # if path is None then we're in '.' and we've done that in (1)
            if path:
                mdir, modName = matchName(path)
                if mdir:
                    return os.path.realpath(path), mdir, modName

    # 3. The plugin directory (mmapath/plugins)
    if tryPlugDir:
        for p in mma.MMA.paths.plugPaths:
            mdir, modName = matchName(p)
            if mdir:
                return p, mdir, modName

    return None, None, None

def hashfile(p):
    """ Calculate a sha-256 hash value on a file. """

    sha256 = hashlib.sha256(b'mmaIsWonderful')

    try:
        f = open(p, 'rb')
    except:
        error("Plugin: Cannot open file '%s'." % p)

    try:
        sha256.update(f.read())
    except:
        error("Plugin: Cannot calculate hash value for plugin file.")
    finally:
        f.close()

    return sha256.hexdigest()

def getPermission(path, name):
    """ Check the file for plugin permissions."""

    global permFile

    if secOverRide:
        return

    # We need somewhere to store this registery. appdirs to the rescue!
    if not permFile:
        cachePath = errorName = mma.MMA.appdirs.user_data_dir('mma', 'Mellowood')
        
        # Can we access the directory? If not, we first try to create
        # a mma directory (we're assuming that HOME/.config, HOME/.local/config,
        # or MMA/lib already exists.
        try:
            os.makedirs(cachePath)
        except OSError:
            if not os.path.isdir(cachePath):
                cachePath = None  # couldn't create directory, that's okay

        if not cachePath:
            warning("Plugin: Can't create/access directory to store permission cache"
                    " file at '%s'. You can still run the plugin, but permissions"
                    " won't be saved." % errorName)

        # We should have a complete path to our storage file.
        if cachePath:
            permFile = os.path.join(cachePath, 'plugins.list')
        else:
            permFile = None

    sha = hashfile(path)
    permlist = {}

    # This is done for each PLUGIN call. Not a big deal, but could be optimized.
    if permFile:
        try:
            f = open(permFile, 'r')
            f.readline()    # Read/discard comment line
            permlist = json.load(f)
            f.close()
        except:
            permlist = {}  # files doesn't exist or can't be read. Ignore.

    # we have a valid permlist dictionary (or an empty one) in memory.
    # if the entries match, all's well

    if path in permlist and  permlist[path] == sha:
        return 

    prettyPrint("PLUGIN: This file is attempting to load the plugin '%s'."
                " As detailed in the documentation a plugin can run arbitrary"
                " Python code and can be dangerous to your system."
                " If you don't understand this, DO NOT"
                " accept loading this plugin!"
                " If you are sure you want to grant permission to load this plugin"
                " press the <y> key followed by <Enter> to register; <o><Enter>"
                " to permit running only once. Any other key combination"
                " will terminate this run." % name)
    a = raw_input("       Your choice: ")

    # Just return. Don't bother to update the permissions table
    if a == 'o':
        return

    # Abort
    if a != 'y':
        error("Plugin: permission refused for loading plugin.")

    # update our permissions dictionary and save it for the next run.

    if permFile:
        permlist[path] = sha
        try:
            f = open(permFile, 'w')
            f.write("### MMA plugin permissions. DO NOT EDIT ###\n")
            json.dump(permlist, f)
            f.close()
        except:
            warning("Plugin: Could not access/create file '%s'. Plugin permission has"
                    " not been saved. You probably have a"
                    " permissions problem on this system." % permFile)

def registerPlugin(p):
    """ Search for the plugin and register any found methods. """

    if p.startswith(prefix):  # user has option to use 
        p = p[1:]

    pdir, mdir, modName = findPlugin(p.upper())

    # pdir - complete path to plug container
    # mdir - path from pdir to the plug. (hello/plugin.py)
    # modName - name for module load (hello.plugin)

    if not pdir:
        error("Plugin: Cannot find a path to %s" % p)

    if modName in plugList:
        warning("Plugin '%s' attempted reload ignored." % modName)
        return

    initMod = os.path.join(pdir, os.path.split(mdir)[0],  "__init__.py")
    if not os.path.exists(initMod):
        error("Plugin needs an empty file '%s'." % initMod)

    if os.stat(initMod).st_size > 0:
        error("Plugin '__init__.py' module must be empty (security concern).")

    plugPath = os.path.join(pdir,mdir)
    plugName =  modName.split('.')[0]

    getPermission(plugPath, plugName)

    plugList.append(modName)
    # insert our plugin path to sys.path so that python's load
    # module function will find our plugs.
    sys.path.insert(0, pdir)

    # load the module. 
    if plugName in sys.modules.keys():
        error("Plugin: the name of the '%s' is already is use by the system. "
              "Most likely MMA has already imported a module by that name. "
              "Please rename the package!" % plugName)
    try:
        e = importlib.import_module(modName, package=None)
    except ImportError as err:
        error("Plugin: Error loading module '%s'. Python is reporting '%s'. "
              "Most likely there is an Import statement in the module which is not working." %
              (modName, str(err)))
        
    # restore old sys.path.
    sys.path.pop(0)

    # The module is now in memory.
    # Find and register the entry points.

    cmdName = prefix + p.upper()

    e.plugInName = {'name': plugName,
                    'dir':  pdir,
                    'path': plugPath,
                    'cmd':  cmdName   }

    try:
        mma.MMA.parse.simpleFuncs[cmdName] = e.run
        simplePlugs.append(cmdName)
        if mma.MMA.debug.debug:
            dPrint("Plugin: %s simple plugin RUN registered." % cmdName.title())
    except:
        pass
    
    try:
        mma.MMA.parse.trackFuncs[cmdName] = e.trackRun
        trackPlugs.append(cmdName)
        if mma.MMA.debug.debug:
            dPrint("Plugin: %s track plugin TrackRun registered." % cmdName.title())
    except:
        pass
    
    try:
        mma.MMA.parse.dataFuncs[cmdName] = e.dataRun
        dataPlugs.append(cmdName)
        if mma.MMA.debug.debug:
            dPrint("Plugin: %s data plugin DataRun registered." % cmdName.title())
    except:
        pass
    
    try:
        plugHelp[cmdName] = e.printUsage
    except:
        plugHelp[cmdName] = None

    return cmdName


def pluginHelp(p):
    """ Called from options to print help message. This is called
        ONLY from the MMA command line ... so no plugs have been
        loaded. We attempt to find the plug, then call its help.
    """

    cmd = registerPlugin(p)

    try:
        plugHelp[cmd]()
    except:
        print("No help message registered for '%s'." % cmd)


def plugin(ln):
    """ Search for, load and install requested plugin(s) """

    global plugsOff, tryLocal, tryDot, tryPlugDir

    if plugsOff:
        error("Plugin loading has been disabled.")

    if not ln:
        error("Plugin requires at least one argument (plugin name).")

    ln, optpair = opt2pair(ln, toupper=True)  # get options
    for cmd, opt in optpair:
        if cmd == 'DISABLE':
            for o in opt.split(','):
                a = o.upper()
                if a == 'ALL':
                    plugsOff = True
                elif a  == 'LOCAL':
                    tryLocal = False
                elif a == 'DOT':
                    tryDot = False
                elif a == 'PLUGDIR':
                    tryPlugDir = False
                else:
                    error("Plugin Disable: '%s' is an unknown or illegal option." % o)

            if mma.MMA.debug.debug:
                if plugsOff:
                    dPrint("Plugin: loading disabled.")
                else:
                    if not tryLocal:
                        dPrint("Plugin: no local plug loading.")
                    if not tryDot:
                        dPrint("Plugin: no current directory plug loading.")
                    if not tryPlugDir:
                        dPrint("Plugin: no plugin directory plug loading.")
        else:
            error("Plugin: '%s' is an unknown command." % cmd)

    # Now load the plugins requested. Note, the args are all in
    # uppercase, but that doesn't matter since the find
    # function checks all possibilities.

    for p in ln:
        registerPlugin(p)





# midiM.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This module contains the MIDI value & command (un)packing routines.

These are necessary to create the big-endian values MIDI expects.

"""

import sys
from struct import pack, unpack
from mma.MMA.common import *
from . import gbl

PY3 = sys.version_info[0] == 3

# this is for the join() in packBtyes(). By predefining
# we only call the bytearry() function once. Yes, it is quicker.
BCAT = bytearray(b'')
ctr={"tuple":0, "str":0, "int":0, "bytes":0}

def packBytes(*args):
    """ Return a packed string of bytes as a single string.
        args - an arbitrary number of strings, ints, lists [int[,int]..]
               or tuples (int[,int]..) 
        return - a single string.

        Note: strings are just concatenated into the string, tuples & ints
              are packed into single value bytes. In the future there
               _may_ be a reason to use struct.pack on string data, but I
              don't see it at this time.

              When calling, make sure tuples/lists are all ints. Having
              a string in a tuple will crash the whole mess!
    """

    ret = []

    for a in args:
        ty = type(a)

        if ty == int:
            ret.append(pack('B', a))  # integers

        # In py2 bytes will also grab string since they are the same
        elif ty in (bytes, bytearray):
            ret.append(a)

        elif ty in (tuple, list):
            ctr["tuple"]+=1
            ret.append(pack('%sB' % len(a), *a))  # tuples/lists

        # In py3 this will handle str. The else will not be done
        # but we'll leave it just in case :)
        elif ty == str:
            if PY3:
                ret.append(pack('%ss' % len(a), a.encode(encoding=gbl.encoding)))
            else:
                ret.append(a)

        else:
            error("Call Bob: Unknown type '%s' for midi string encode." % ty)

    return BCAT.join(ret)



def intToWord(x):
    """ Convert INT to a 2 byte MSB LSB value. Used in MIDI headers. """

    return pack('>H', x)


def intTo3Byte(x):
    """ Convert INT to a 3 byte MSB...LSB value. """

    return pack('>L', int(x))[1:]


def byte3ToInt(x):

    x = pack('1B', 0) + x
    return unpack('>L', x)[0]


def intToLong(x):
    """ Convert INT to a 4 byte MSB...LSB value. """

    return pack('>L', x)


def intToVarNumber(x):
    """ Convert INT to a variable length MIDI value. """
    
    lst = [ x & 0x7f ]
    while 1:
        x = x >> 7
        if x:
            lst.insert(0, (x & 0x7f) | 0x80)
        else:
            return pack("%sB" % len(lst), *lst)


def intTo14(x):
    """ Convert INT to a 2 byte 14 bit number

        Value to be converted must be 0...16383. A
        ValueError is raised if out of range.
    """

    if x < 0 or x > 0x3fff:
        raise ValueError

    return pack("2B", 0x007f & x, x >> 7)



# debug.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

Some silly functions used by the debug code.


"""

# the Lxxx values are the previous settings, used for LASTDEBUG macro

debug          =     Ldebug         = 0
pshow          =     Lpshow         = 0
seqshow        =     Lseqshow       = 0
showrun        =     Lshowrun       = 0
noWarn         =     LnoWarn        = 0
noOutput       =     LnoOutput      = 0
showExpand     =     LshowExpand    = 0
showFilenames  =     LshowFilenames = 0
chshow         =     Lchshow        = 0

plecShow       =     LplecShow  = 0  # not a command line setting
rmShow         =     LrmShow    = 0  # not command
gvShow         =     LgvShow    = 0

def cmdLineDebug(o):
    """ Set a command line debug option. Called from options.py """

    global debug, Ldebug, showFilenames, LshowFilenames, \
        pshow, Lpshow, seqshow, Lseqshow, showrun, Lshowrun, \
        noWarn, LnoWarn, noOutput, LnoOutput, showExpand, \
        LshowExpand, chshow, Lchshow 
    
    if o == 'd':
        debug = Ldebug = 1

    elif o == 'o':
        showFilenames = LshowFilenames = 1

    elif o == 'p':
        pshow = Lpshow = 1

    elif o == 's':
        seqshow = Lseqshow = 1

    elif o == 'r':
        showrun = Lshowrun = 1

    elif o == 'w':
        noWarn = LnoWarn = 1

    elif o == 'n':
        noOutput = LnoOutput = 1

    elif o == 'e':
        showExpand = LshowExpand = 1

    elif o == 'c':
        chshow = Lchshow = 1
    
def setDebug(ln):
    """ Set debugging options dynamically. """

    # This needs to be here to avoid circular import problem
    from mma.MMA.common import opt2pair
    from mma.MMA.common import dPrint
    
    global Ldebug, debug, LshowFilenames, showFilenames, \
        Lpshow, pshow, Lseqshow, seqshow, Lshowrun, showrun, \
        LnoWarn, noWarn, LnoOutput, noOutput, LshowExpand, showExpand, \
        Lchshow, chshow, LplecShow, plecShow, LrmShow, rmShow, \
        LgvShow, gvShow
    
    msg = ("Debug: Use MODE=On/Off where MODE is one or more of "
           "DEBUG, FILENAMES, PATTERNS, SEQUENCE, GROOVE, "
           "RUNTIME, WARNINGS, EXPAND, ROMAN or PLECTRUM.")

    if not len(ln):
        from mma.MMA.common import error
        error(msg)

    # save current flags

    Ldebug = debug
    LshowFilenames = showFilenames
    Lpshow = pshow
    Lseqshow = seqshow
    Lshowrun = showrun
    LnoWarn = noWarn
    LnoOutput = noOutput
    LshowExpand = showExpand
    Lchshow = chshow
    LplecShow = plecShow
    LrmShow = rmShow
    LgvShow = gvShow

    ln, opts = opt2pair(ln, 1)
    if ln:
        error("Each debug option must be a opt=value pair.")

    for cmd, val in opts:
        if val == 'ON' or val == '1':
            val = 1
        elif val == 'OFF' or val == '0':
            val = 0
        else:
            error("Debug: %s needs ON, 1, OFF, or 0 arg." % cmd)

        if cmd == 'DEBUG':
            debug = val
            if debug:
                dPrint("Debug=%s." % val)

        elif cmd == 'FILENAMES':
            showFilenames = val
            if debug:
                dPrint("ShowFilenames=%s." % val)

        elif cmd == 'PATTERNS':
            pshow = val
            if debug:
                dPrint("Pattern display=%s." % val)

        elif cmd == 'SEQUENCE':
            seqshow = val
            if debug:
                dPrint("Sequence display=%s." % val)

        elif cmd == 'RUNTIME':
            showrun = val
            if debug:
                dPrint("Runtime display=%s." % val)

        elif cmd == 'WARNINGS':
            noWarn = not(val)
            if debug:
                dPrint("Warning display=%s" % val)

        elif cmd == 'EXPAND':
            showExpand = val
            if debug:
                dPrint("Expand display=%s." % val)

        elif cmd == 'ROMAN':
            rmShow = val
            if debug:
                dPrint("Roman numeral chords/slash display=%s" % val)

        elif cmd == 'GROOVE':
            gvShow = val
            if debug:
                dPrint("Groove re-define display=%s" % val)

        elif cmd == 'PLECTRUM':
            plecShow = val
            if debug:
                dPrint("Plectrum display=%s" % val)

        else:
            from mma.MMA.common import error
            error(msg)

            
def getFlags():
    """ Returns current values of debug flags in a string.
        Used by macro.py to expand $_Debug. 
    """    
    
    return "Debug=%s  Filenames=%s Patterns=%s " \
        "Sequence=%s Runtime=%s Warnings=%s Expand=%s " \
        "Roman=%s Plectrum=%s Groove=%s" % \
           (debug, showFilenames, pshow, seqshow, showrun,
            int(not noWarn), showExpand, rmShow, plecShow, gvShow)

def getLFlags():
    """ Returns last set values of debug flags in a string.
        Used by macro.py to expand %_LastDebug. 
    """
    
    return "Debug=%s  Filenames=%s Patterns=%s " \
                "Sequence=%s Runtime=%s Warnings=%s Expand=%s " \
                "Roman=%s Plectrum=%s Groove=%s" % \
                (Ldebug, LshowFilenames, Lpshow, Lseqshow, Lshowrun,
                 int(not LnoWarn), LshowExpand, LrmShow, LplecShow, LgvShow)

def trackSet(track, func):
    """ Print a debug message for functions which set a track setting. 

        track - name of the track (Chord-Sus ... )
        name - the name of the function (Strum, Compress ...)

        By using the macro formatting functions we get consistent output 
         between debug and macro expansion.
    """

    # Need to do this way to avoid circular import problem
    from mma.MMA.macro import macros
    from mma.MMA.common import dPrint
    
    dPrint("Set %s %s: %s" % (track, func,
            macros.sysvar("%s_%s" % (track, func.upper()))))







# chords.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

from mma.MMA.common import *
from mma.MMA.chordtable import chordlist
import mma.MMA.roman
from mma.MMA.keysig import keySig  # needed for voicing mode keycenter()
import mma.MMA.debug

import copy

slashPrinted = []  # for slash chord error message

####################################################
# Convert a roman numeral chord to standard notation


def defChord(ln):
    """ Add a new chord type to the chords{} dict. """

    emsg = "DefChord needs NAME (NOTES) (SCALE)"

    # At this point ln is a list. The first item should be
    # the new chord type name.

    if not len(ln):
        error(emsg)
    name = ln.pop(0)
    if name in chordlist.keys():
        warning("Redefining chordtype '%s'" % name)

    if '/' in name:
        error("A slash in not permitted in chord type name")

    if '>' in name:
        error("A '>' in not permitted in chord type name")

    ln = pextract(''.join(ln), '(', ')')

    if ln[0] or len(ln[1]) != 2:
        error(emsg)

    notes = ln[1][0].split(',')
    if len(notes) < 2 or len(notes) > 8:
        error("There must be 2..8 notes in a chord, not '%s'" % len(notes))

    for i, v in enumerate(notes):
        v = stoi(v, "Note offsets in chord must be integers, not '%s'" % v)
        if v < 0 or v > 24:
            error("Note offsets in chord must be 0..24, not '%s'" % v)
        notes[i] = v

    scale = ln[1][1].split(',')
    if len(scale) != 7:
        error("There must be 7 offsets in chord scale, not '%s'" % len(scale))

    for i, v in enumerate(scale):
        v = stoi(v, "Scale offsets in chord must be integers, not '%s'" % v)
        if v < 0 or v > 24:
            error("Scale offsets in chord must be 0..24, not '%s'" % v)
        scale[i] = v

    chordlist[name] = (notes, scale, "User Defined")

    if mma.MMA.debug.debug:
        dPrint("ChordType '%s', %s" % (name, chordlist[name]))


def printChord(ln):
    """ Display the note/scale/def for chord(s). """

    for c in ln:
        try:
            print("%s: %s %s  %s" % 
                  (c, tuple(chordlist[c][0]),
                   tuple(chordlist[c][1]), chordlist[c][2]))
        except KeyError:
            error("Chord '%s' is unknown. Use only the chord type (m, M, M7, ...)." % c)


""" Table of chord adjustment factors. Since the initial chord is based
    on a C scale, we need to shift the chord for different degrees. Note,
    that with C as a midpoint we shift left for G/A/B and right for D/E/F.

    Should the shifts take in account the current key signature?
"""

cdAdjust = {
    'Gb': -6,
    'G' : -5,
    'G#': -4, 'Ab':-4,
    'A' : -3,
    'A#': -2, 'Bb':-2,
    'B' : -1, 'Cb':-1,
    'B#':  0, 'C' : 0,
    'C#': 1, 'Db': 1,
    'D' : 2,
    'D#': 3, 'Eb': 3,
    'E' : 4, 'Fb': 4,
    'E#': 5, 'F' : 5,
    'F#': 6 }

cdAdjustOrig = copy.copy(cdAdjust)

def chordAdjust(ln):
    """ Adjust the chord point up/down one octave. """

    global cdAdjust

    args, ln = opt2pair(ln)

    if not ln and not args:
        error("ChordAdjust: Needs at least one argument.")

    for a in args:
        if a.upper() == 'RESET':
            cdAdjust = copy.copy(cdAdjustOrig)
        
        else:
            error("ChordAdjust: %s is not a valid argument." % a)

    for p, octave in ln:
        octave = stoi(octave, "ChordAdjust: expecting integer, not '%s'" % octave)
        if octave not in (-1, 0, 1):
            error("ChordAdjust: '%s' is not a valid octave. Use 1, 0 or -1" % octave)

        for pitch in p.split(','):
            if pitch not in cdAdjust:
                error("ChordAdjust: '%s' is not a valid pitch" % pitch)

            if octave == 0:
                cdAdjust[pitch] = cdAdjustOrig[pitch]
            else:
                cdAdjust[pitch] = cdAdjustOrig[pitch] + (octave * 12)
       


###############################
# Chord creation/manipulation #
###############################


class ChordNotes:
    """ The Chord class creates and manipulates chords for mma.MMA. The
    class is initialized with a call with the chord name. Eg:

    ch = ChordNotes("Am")

    The following methods and variables are defined:

    name - original chord name without leading +/- and trail > etc.

    noteList  - the notes in the chord as a list. The "Am"
        would be [9, 12, 16].

    noteListLen     - length of noteList.

    tonic       - the tonic of the chord ("Am" would be "A").

    chordType  - the type of chord ("Am" would be "m").

    rootNote   - the root note of the chord ("Am" would be a 9).

    bnoteList  - the original chord notes, bypassing any
                     invert(), etc. mangling.

    scaleList  - a 7 note list representing a scale similar to
                     the chord.

    reset() - resets noteList to the original chord notes.
              This is useful to restore the original after
              chord note mangling by invert(), etc. without having to
              create a new chord object.


    invert(n) - Inverts a chord by 'n'. This is done inplace and
                returns None. 'n' can have any integer value, but -1 and 1
                are most common. The order of the notes is not changed. Eg:

                ch=Chord('Am')
                ch.noteList == [9, 12, 16]
                ch.invert(1)
                ch.noteList     = [21, 12, 16]

    compress() - Compresses the range of a chord to a single octave. This is
                 done inplace and returns None. Eg:

                 ch=Chord("A13")
                 ch.noteList == [1, 5, 8, 11, 21]
                 ch.compress()
                 ch.noteList == [1, 5, 8, 11, 10 ]


    limit(n) -    Limits the range of the chord 'n' notes. Done inplace
                  and returns None. Eg:

                  ch=Chord("CM711")
                  ch.noteList == [0, 4, 7, 11, 15, 18]
                  ch.limit(4)
                  ch.noteList ==    [0, 4, 7, 11]


    """

    #################
    ### Functions ###
    #################

    def __init__(self, name):
        """ Create a chord object. Pass the chord name as the only arg.

        NOTE: Chord names ARE case-sensitive!

            The chord NAME at this point is something like 'Cm' or 'A#7'.
            Split off the tonic and the type.
            If the 2nd char is '#' or 'b' we have a 2 char tonic,
            otherwise, it's the first char only.

            A chord can start with a single '+' or '-'. This moves
            the entire chord and scale up/down an octave.

            Note pythonic trick: By using ranges like [1:2] we
            avoid runtime errors on too-short strings. If a 1 char
            string, name[1] is an error; name[1:2] just returns None.

            Further note: I have tried to enable caching of the generated
            chords, but found no speed difference. So, to make life simpler
            I've decided to generate a new object each time.

        """

        slash = None
        wmessage = ''   # slash warning msg, builder needed for debug.rmShow
        octave = 0
        inversion = 0
        polychord = None
        startingName = name

        if name == 'z':
            self.tonic = self.chordType = None
            self.noteListLen = 0
            self.notesList = self.bnoteList = []
            return

        if '|' in name:
            if name.count('|') > 1:
                error("Polychord marker '|' can only be used once.")

            name, polychord = name.split('|')

        if '/' in name and '>' in name:
            error("You cannot use both an inversion and a slash in the same chord")

        if ':' in name:
            name, barre = name.split(':', 1)
            barre = stoi(barre, "Expecting integer after ':'")
            if barre < -127 or barre > 127:
                error("Chord barres limited to -127 to 127 (more than about +/-20 is silly)")
        else:
            barre = 0

        if '>' in name:
            name, inversion = name.split('>', 1)
            inversion = stoi(inversion, "Expecting integer after '>'")
            if inversion < -5 or inversion > 5:
                error("Chord inversions limited to -5 to 5 (more seems silly)")

        while name[0:1] in ['-', '+']:
            if name[0] == '-':
                octave -= 12
            else:
                octave += 12
            if abs(octave) > 96:
                error("Too many octave adjustments in '%s'." % startingName)
            name = name[1:]

        # we have just the name part. Save 'origname' for debug print

        self.name = name = name.replace('&', 'b')

        if not name:
            error("Empty chord name generated from '%s'." % startingName)

        # Strip off the slash part of the chord. Use later
        # to do proper inversion.

        if name.find('/') > 0:
            name, slash = name.split('/')

        # Name stars with roman (I, V, i, v) then assume
        # rest of name is valid roman. Convert to "standard"
        if name[0] in ("I", "V", "i", "v"):
            n = name
            name = mma.MMA.roman.convert(name)

        # Split chord name (A,B..) and type (minor, dim)
        if name[1:2] in ('#b'):
            tonic = name[0:2]
            ctype = name[2:]
        else:
            tonic = name[0:1]
            ctype = name[1:]

        if not ctype:        # If no type, make it a Major
            ctype = 'M'

        # Now we get the notes for this chord. Just based on the type.
        # Adjust it up/down for the tonic (c,d...)
        try:
            notes = chordlist[ctype][0]
            adj = cdAdjust[tonic] + octave
        except KeyError:
            error("Illegal/Unknown chord name: '%s'" % name)

        self.noteList = [x + adj for x in notes]
        self.bnoteList = tuple(self.noteList)
        self.scaleList = tuple([x + adj for x in chordlist[ctype][1]])
        self.chordType = ctype
        self.tonic = tonic
        self.rootNote = self.noteList[0]
        self.barre = barre

        self.noteListLen = len(self.noteList)

        # Inversion

        if inversion:
            self.invert(inversion)
            self.bnoteList = tuple(self.noteList)

        # Do inversions if there is a valid slash notation.

        if slash:   # convert Roman or Arabic to name of note from chord scale
            if slash[0] in ('I', 'i', 'V', 'v') or slash[0].isdigit():
                n = mma.MMA.roman.rvalue(slash)
                n = self.scaleList[n] % 12   # midi value 

                slash = ('C', 'C#', 'D', 'D#', 'E', 'F',
                         'F#', 'G', 'G#', 'A', 'A#', 'B')[n]
            try:
                r = cdAdjust[slash]    # r = -6 to 6
            except KeyError:
                error("The note '%s' in the slash chord is unknown" % slash)

            # If the slash note is in the chord we invert
            # the chord so the slash note is in root position.
            # NOTE: 'r' is the value of the slash note. NOTE:
            # this is a true rotate, the self.invert() routine
            # fakes it by changing note octaves.

            # Check the slash note against the chord notes
            c_roted = 0
            s = self.noteList
            for octave in [0, 12, 24]:
                if r+octave in s:
                    rot = s.index(r+octave)
                    for i in range(rot):
                        s.append(s.pop(0)+12)
                    if s[0] >= 12:
                        s = [v-12 for v in s]
                    self.noteList = s
                    self.bnoteList = tuple(s)
                    self.rootNote = self.noteList[0]
                    c_roted = 1
                    break

            # Check against the scale notes. If the note is in
            # the scale we rotate the scale to force the slash note
            # to the root position. 
            s_roted = 0
            s = list(self.scaleList)
            for octave in [0, 12, 24]:
                if r+octave in s:
                    rot = s.index(r+octave)
                    for i in range(rot):
                        s.append(s.pop(0)+12)
                    if s[0] >= 12:
                        s = [ v-12 for v in s ]
                    self.scaleList = tuple(s)
                    s_roted = 1
                    break

            # Slash note is not in the chord or scale. This is
            # not an error ... but the note is ignored. So, print
            # a message and list out alternate chords to use.
            if not c_roted and not s_roted:
                wmessage = "The slash chord note '%s' not in chord or scale '%s'" % \
                         (slash, name)

                t = "%s/%s" % (name, slash )
                if t not in slashPrinted:
                    note = r % 12
                    ll = []
                    for c in chordlist:
                        if ord(c[0]) > 128:  # skip dim. symbol
                            continue
                        # We need 'adj' to convert the chords from "C" to the current tonic.
                        adj = cdAdjust[self.tonic]
                        if note in  [(x % 12) + adj for x in chordlist[c][0]]:
                            ll.append("%s%s" % (self.tonic, c))
                    if ll:
                        wmessage += "\nChords with '%s': %s" % (slash, ' '.join(sorted(ll)))

                    slashPrinted.append(t)  # only print this chord/slash once
                if not mma.MMA.debug.rmShow:
                    warning(wmessage)

        if polychord:
            ctable = ChordNotes(polychord)
            self.noteList.extend(ctable.noteList)
            self.noteList = list(set(self.noteList))
            self.noteList.sort()
            if len(self.noteList) > 8:
                i = len(self.noteList)
                self.noteList = self.noteList[:8]
                warning("Polychord %s|%s has been truncated from %d to 8 notes." % \
                        (name, polychord, i))
            self.noteListLen = len(self.noteList)
            self.bnoteList = tuple(self.noteList)

        if mma.MMA.debug.rmShow:  # Display roman debug (Debug=Roman)
            if slash:
                a = '/' + slash
            else:
                a = ''
            if wmessage:
                a += '   ' + wmessage
            dPrint(" %03s] %-09s -> %s%s" % (gbl.lineno, startingName, name, a))
        
    def reset(self):
        """ Restores notes array to original, undoes mangling. """

        self.noteList = list(self.bnoteList[:])
        self.noteListLen = len(self.noteList)


    def invert(self, n):
        """ Apply an inversion to a chord.

        This does not reorder any notes, which means that the root note of
        the chord reminds in position 0. We just find that highest/lowest
        notes in the chord and adjust their octave.

        NOTE: Done on the existing list of notes. Returns None.
        """

        if n:
            c = self.noteList[:]
            while n > 0:        # Rotate up by adding 12 to lowest note
                n -= 1
                c[c.index(min(c))] += 12

            while n < 0:        # Rotate down, subtract 12 from highest note
                n += 1
                c[c.index(max(c))] -= 12

            self.noteList = c

        return None



    def compress(self):
        """ Compress a chord to one ocatve. """

        # Get the max value in the chord list. This is the
        # lowest note in the chord + 12. Note: use the unmodifed value bnoteList!
        mx = min(self.bnoteList) + 12
        c=[]
        for n in self.noteList:
            if n > mx:
                n -= 12
            c.append(n)

        self.noteList = c
        return None



    def limit(self, n):
        """ Limit the number of notes in a chord. """

        if n < self.noteListLen:
            self.noteList =     self.noteList[:n]
            self.noteListLen = len(self.noteList)

        return None

    def keycenter(self):
        """ Rotate the chord notes until the tonic is close to
            the keysig value. Note that keysig is set C=0, D=2, etc
            which matches the note values in the noteList.
        """

        key = keySig.keyNoteValue
        notes = self.noteList  # just a shorter variable name

        nbelow=0
        nabove=0

        for n in notes:
            if n<key: nbelow+=1
            elif n>key: nabove+=1

        if nbelow>nabove:
            for a in range(nbelow-nabove-1): # this is 0,1 ... start of list
                notes[a] += 12

        elif nbelow<nabove:
            for a in range(1,nabove-nbelow):   # this should give 1,2 
                notes[-a] -= 12                # and the - means end of list

        return None

    def key2center(self):
        """ This is the same as 'key', but the 9th, 11th, etc. are not modified.
            We do this by changing the note list so that 9th, etc. are deleted,
            the 'key' function is called and the deleted notes are restored.
        """
        
        orig = self.noteList[:]

        saved=[]
        self.noteList = []
        root = orig[0]
        for n in orig: 
            if n >= root+12:
                saved.append(n)
            else:
                self.noteList.append(n)
        self.keycenter()
        self.noteList.extend(saved)

        return None

    def rootkey(self):
        """ Modify the chord so that the notes are all close to, but above, the 
            root note (from the key sig).
        """
        
        key = keySig.keyNoteValue
        new = []

        for n in self.noteList: 
            while n<0:
                n+=12
            while n>12:
                n-=12

            if n-12 >= key:
                new.append(n-12)
            else:
                new.append(n)

        self.noteList = new

    def center1(self, lastChord):
        """ Descriptive comment needed here!!!! """

        def minDistToLast(x, lastChord):
            dist=99
            for j in range(len(lastChord)):
                if abs(x-lastChord[j]) < abs(dist):
                    dist = x-lastChord[j]
            return dist

        def sign(x):
            if (x>0):
                return 1
            elif (x<0):
                return -1
            else:
                return 0

        # Only change what needs to be changed compared to the last chord
        # (leave notes where they are if they are in the new chord as well).

        if lastChord:
            ch=self.noteList

            for i in range(len(ch)):

                # minimize distance to last chord

                oldDist = minDistToLast(ch[i], lastChord)
                while abs(minDistToLast(ch[i] - sign(oldDist)*12,
                                lastChord)) < abs(oldDist):
                    ch[i] -= 12* sign(oldDist)
                    oldDist = minDistToLast(ch[i], lastChord)

        return None

    def center2(self, centerNote, noteRange):
        """ Need COMMENT """

        ch=self.noteList
        for i,v in enumerate(ch):

            dist = v - centerNote
            if dist < -noteRange:
                ch[i] = v + 12 * ( abs(dist) // 12+1 )
            if dist > noteRange:
                ch[i] = v - 12 * ( abs(dist) // 12+1 )


        return None



######## End of Chord class #####







from .mma import *
from .MMA import *

# MMAdir and platform for package
from os import path
import platform
global MMAdir, platform_so
MMAdir = path.join(path.dirname(__file__))
platform_so = platform.system()




# parse.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


This module does all file parsing. Most commands
are passed to the track classes; however, things
like TIME, SEQRND, etc. which just set global flags
are completely handled here.

"""

import random

from . import gbl
from mma.MMA.common import *

import mma.MMA.debug
import mma.MMA.notelen
import mma.MMA.chords
import mma.MMA.file
import mma.MMA.midi
import mma.MMA.midifuncs
import mma.MMA.midiIn
import mma.MMA.midinote
import mma.MMA.grooves
import mma.MMA.docs
import mma.MMA.auto
import mma.MMA.func
import mma.MMA.translate
import mma.MMA.patSolo
import mma.MMA.mdefine
import mma.MMA.volume
import mma.MMA.seqrnd
import mma.MMA.patch
import mma.MMA.paths
import mma.MMA.player
import mma.MMA.sequence
import mma.MMA.swing
import mma.MMA.sync
import mma.MMA.truncate
import mma.MMA.ornament
import mma.MMA.trigger
import mma.MMA.tempo
import mma.MMA.tweaks
import mma.MMA.options
import mma.MMA.rpitch
import mma.MMA.regplug
import mma.MMA.after

from mma.MMA.timesig import timeSig
from mma.MMA.parseCL import parseChordLine
from mma.MMA.lyric import lyric
from mma.MMA.macro import macros
from mma.MMA.alloc import trackAlloc
from mma.MMA.keysig import keySig

beginData = []      # Current data set by a BEGIN statement
beginPoints = []    # since BEGINs can be nested, we need ptrs for backing out of BEGINs


########################################
# File processing. Mostly jumps to pats
########################################

def parseFile(n):
    """ Open and process a file. Errors exit. """

    fp = gbl.inpath

    f = mma.MMA.file.ReadFile(n)

    parse(f)
    gbl.inpath = fp

    if mma.MMA.debug.debug:
        dPrint("File '%s' closed." % n)


    
                
def parse(inpath):
    """ Process a mma input file. """

    global beginData, lastChord
    
    gbl.inpath = inpath
    curline = None

    while 1:
        mma.MMA.after.check()
        
        curline = inpath.read()

        if curline is None:   # eof, exit parser
            break

        l = macros.expand(curline)
        if not l:
            continue

        """ Handle BEGIN and END here. This is outside of the Repeat/End
            and variable expand loops so SHOULD be pretty bullet proof.
            Note that the beginData stuff is global to this module ... the
            Include/Use directives check to make sure we're not doing that
            inside a Begin/End.

            beginData[] is a list which we append to as more Begins are
            encountered.

            The placement here is pretty deliberate. Variable expand comes
            later so you can't macroize BEGIN ... I think this makes sense.

            The tests for 'begin', 'end' and the appending of the current
            begin[] stuff have to be here, in this order.
        """

        action = l[0].upper()      # 1st arg in line

        if action == 'BEGIN':
            if not l[1:]:
                error("Use: Begin STUFF")
            beginPoints.append(len(beginData))
            beginData.extend(l[1:])
            continue

        if action == 'END':
            if len(l) > 1:
                error("No arguments permitted for END")
            if not beginData:
                error("No 'BEGIN' for 'END'")
            beginData = beginData[:beginPoints.pop(-1)]
            continue

        if beginData:
            l = beginData + l
            action = l[0].upper()

        if mma.MMA.debug.showExpand and action != 'REPEAT':
            dPrint(l)

        if action in simpleFuncs:
            simpleFuncs[action](l[1:])
            continue            

        """ We have several possibilities ...
            1. The command is a valid assigned track name,
            2. The command is a valid track name, but needs to be
               dynamically allocated,
            3. It's really a chord action
        """

        if not action in gbl.tnames:  # no track allocated?
            trackAlloc(action, 0)     # Try to create. Always returns.

        if action in gbl.tnames:     # BASS/DRUM/APEGGIO/CHORD
            name = action
            if len(l) < 2:
                error("Expecting argument after '%s'" % name)
            action = l[1].upper()

            # Got trackname and action
            if action in trackFuncs:  # perfect, execute
                trackFuncs[action](name, l[2:])
                continue

            elif action in simpleFuncs:  # opps, not track func
                error("%s is not a track function. Use global form." % action)
            else:  # opps, not any kind of func
                error("%s is not a %s track function." % (action, name))


        ### Gotta be a chord data line!

        """ A data line can have an optional bar number at the start
            of the line. Makes debugging input easier. The next
            block strips leading integers off the line. Note that
            a line number on a line by itself it okay.
        """

        if action.isdigit():   # isdigit() matches '1', '1234' but not '1a'!
            gbl.barLabel = l[0].lstrip('0')
            l = l[1:]
            if not l:        # ignore empty lines
                continue
        else:
            gbl.barLabel = ''

        ##  A bar can have an optional repeat count. This must
        ##  be at the end of bar in the form '* xx'.
        if len(l) > 1 and l[-2] == '*':
            rptcount = stoi(l[-1], "Expecting integer after '*'")
            l = l[:-2]
        else:
            rptcount = 1

        # Dataplugins all start with '@'. Code is in the plugin code.
        # A data plugin modifies the existing data line and returns it.
        if l[0].startswith('@'):
            p = l[0].upper()
            if p not in dataFuncs:
                error("Unknown data plugin '%s' called." % p)
            l = dataFuncs[p](l[1:])

        # Extract solo(s) from line ... this is anything in {}s.
        # The solo data is pushed into RIFFs and discarded from
        # the current line.
        l = ' '.join(l)
        l = mma.MMA.patSolo.extractSolo(l, rptcount)

        # set lyrics from [stuff] in the current line.
        # NOTE: lyric.extract() inserts previously created
        #       data from LYRICS SET and inserts the chord names
        #       if that flag is active.
        l, lyrics = lyric.extract(l, rptcount)
        l = l.split()

        # At this point we have only chord info. A number
        # of sanity checks are made:
        #   1. Make sure there is some chord data,
        #   2. Ensure the correct number of chords.
        if not l:
            error("Expecting music (chord) data. Even lines with "
                  "lyrics or solos still need a chord. If you "
                  "don't want a chord use 'z'.")

        # We now have a chord line. It'll look something like:
        #
        #      ['Cm', '/', 'z', 'F#@4.5'] or ['/' 'C@3' ]
        #
        #    For each bar we create a list of CTables, one for each
        #    chord in the line. Each entry has the start/end (in beats), chordname, etc.
        #

        ctable = parseChordLine(l)   # parse the chord line

        # Create MIDI data for the bar

        for rpt in range(rptcount):   # for each bar in the repeat count ( Cm * 3)

            """ Handle global (de)cresc by popping a new volume off stack. """

            if mma.MMA.volume.futureVol:
                mma.MMA.volume.volume = mma.MMA.volume.futureVol.pop(0)
            if mma.MMA.volume.futureVol:
                mma.MMA.volume.nextVolume = mma.MMA.volume.futureVol[0]
            else:
                mma.MMA.volume.nextVolume = None

            # Set up for rnd seq. This may set the current seq point.
            # If return is >=0 then we're doing track rnd.

            rsq, seqlist = mma.MMA.seqrnd.setseq()

            """ Process each track. It is important that the track classes
                are written so that the ctable passed to them IS NOT MODIFIED.
                This applies especially to chords. If the track class changes
                the chord, then the function called MUST restore it before returning!!!
            """

            for a in gbl.tnames.values():
                if rsq >= 0:
                    seqSave = gbl.seqCount
                    if a.name in seqlist:   # for seqrnd with tracklist
                        gbl.seqCount = rsq
                a.bar(ctable)    # process entire bar!

                if rsq >= 0:   # for track rnd
                    gbl.seqCount = seqSave

            # Adjust counters

            """ After processsing each bar we update a dictionary of bar
                pointers. This table is used when the MIDI data is written
                when -b or -B is set to limit output.
            """

            if mma.MMA.truncate.length:
                nextOffset = mma.MMA.truncate.length
                mma.MMA.truncate.countDown()
            else:
                nextOffset = gbl.barLen

            # barPtrs is used by the -B/b options to strip unwanted sections. 
            gbl.barPtrs[gbl.barNum + 1] = [gbl.barLabel, gbl.tickOffset,
                                           gbl.tickOffset + nextOffset - 1]

            gbl.totTime += float(nextOffset / gbl.BperQ) / gbl.tempo

            gbl.tickOffset += nextOffset

            if gbl.printProcessed:
                if gbl.barLabel:
                    gbl.barLabels.append(gbl.barLabel)
                else:
                    gbl.barLabels.append('?')

            gbl.barNum += 1
            gbl.seqCount = (gbl.seqCount + 1) % gbl.seqSize

            if gbl.barNum > gbl.maxBars:
                error("Capacity exceeded. Maxbar setting is %s. Use -m option"
                      % gbl.maxBars)

            mma.MMA.grooves.nextGroove()   # using groove list? Advance.

            # Enabled with the -r command line option

            if mma.MMA.debug.showrun:
                if lyrics:       # we print lyric as a list 
                    ly = lyrics  # with the []s
                else:
                    ly = ''      # no lyric
                dPrint("%3d: %s %s" % (gbl.barNum, ' '.join(l), ly))

            # if repeat count is set with dupchord we push
            # the chord back and get lyric.extract to add the
            # chord to the midi file again. A real lyric is
            # just ignored ... 2 reasons: the lyric is mangled and
            # and it makes sense to only have it once!
            if rpt and lyric.dupchords:
                _,lyrics = lyric.extract(' '.join(l), 0)

            
##################################################################

def allTracks(ln):
    """ Apply command to all specified tracks or track types. """

    types1 = ('BASS', 'CHORD', 'ARPEGGIO', 'SCALE', 'DRUM', 'WALK', 'PLECTRUM')
    types2 = ('MELODY', 'SOLO', 'ARIA')
    allTypes = types1 + types2

    ttypes = []

    if len(ln) < 1:
        error("AllTracks: Requires arguments: [Track | Track-Name] command.")

    i = 0
    while i < len(ln) and ln[i].upper() in allTypes:
        ttypes.append(ln[i].upper())
        i += 1

    if ttypes == []:
        ttypes = types1

    if i >= len(ln):
        error("AllTracks: A command is required after the Track or Track-Name.")

    cmd = ln[i].upper()
    args = i + 1

    if not cmd in trackFuncs:
        error("AllTracks: command '%s' doen't exist" % cmd)

    for n in gbl.tnames:
        if not gbl.tnames[n].vtype in ttypes:
            continue

        trackFuncs[cmd](n, ln[args:])


#######################################
# Do-nothing functions

def comment(ln):
    pass


def repeatend(ln):
    error("Repeatend/EndRepeat without Repeat")


def repeatending(ln):
    error("Repeatending without Repeat")


def endmset(ln):
    error("EndMset/MSetEnd without If")


def ifend(ln):
    error("ENDIF without IF")


def ifelse(ln):
    error("ELSE without IF")


#######################################
# Repeat/jumps


def repeat(ln):
    """ Repeat/RepeatEnd/RepeatEnding.

        Read input until a RepeatEnd is found. The entire
        chunk is pushed back into the input stream the
        correct number of times. This accounts for endings and
        nested repeats.
    """

    def repeatChunk():
        q = []
        qnum = []
        nesting = 0

        while 1:
            l = gbl.inpath.read()

            if not l:
                error("EOF encountered processing Repeat")

            act = l[0].upper()

            if act == 'REPEAT':
                nesting += 1

            elif act in ('REPEATEND', 'ENDREPEAT') and nesting:
                nesting -= 1

            elif act == 'REPEATENDING' and nesting:
                pass

            elif act in ('REPEATEND', 'ENDREPEAT', 'REPEATENDING'):
                return (q, qnum, act, l[1:])

            q.append(l)
            qnum.append(gbl.lineno)

    stack = []
    stacknum = []
    main = []
    mainnum = []
    ending = 0

    if ln:
        error("REPEAT takes no arguments")

    main, mainnum, act, l = repeatChunk()

    while 1:
        if act in ('REPEATEND', 'ENDREPEAT'):
            if l:
                l = macros.expand(l)
                if len(l) == 2:
                    l = [x.upper() for x in l]
                    if 'NOWARN' in l:
                        l.remove('NOWARN')
                        warn = 0
                else:
                    warn = 1

                if len(l) != 1:
                    error("%s: Use [NoWarn] Count" % act)

                count = stoi(l[0], "%s takes an integer arg" % act)

                if count == 2 and warn:
                    warning("%s count of 2 duplicates default. Did you mean 3 or more?" % act)

                elif count == 1 and warn:
                    warning("%s count of 1 means NO REPEAT" % act)

                elif count == 0 and warn:
                    warning("%s count of 0, Skipping entire repeated section" % act)

                elif count < 0:
                    error("%s count must be 0 or greater" % act)

                elif count > 10 and warn:
                    warning("%s is a large value for %s" % (count, act))

            else:
                count = 2

            if not ending:
                count += 1
            for c in range(count - 1):
                stack.extend(main)
                stacknum.extend(mainnum)
            gbl.inpath.push(stack, stacknum)
            break

        elif act == 'REPEATENDING':
            ending = 1

            if l:
                l = macros.expand(l)
                if len(l) == 2:
                    l = [x.upper() for x in l]
                    if 'NOWARN' in l:
                        l.remove('NOWARN')
                        warn = 0
                else:
                    warn = 1

                if len(l) != 1:
                    error("REPEATENDING: Use [NoWarn] Count")

                count = stoi(l[0], "RepeatEnding takes an integer arg")

                if count < 0:
                    error("RepeatEnding count must be postive, not '%s'" % count)

                elif count == 0 and warn:
                    warning("RepeatEnding count of 0, skipping section")

                elif count == 1 and warn:
                    warning("RepeatEnding count of 1 duplicates default")

                elif count > 10 and warn:
                    warning("%s is a large value for RepeatEnding" % count)
            else:
                count = 1

            rpt, rptnum, act, l = repeatChunk()

            for c in range(count):
                stack.extend(main)
                stacknum.extend(mainnum)
                stack.extend(rpt)
                stacknum.extend(rptnum)

        else:
            error("Unexpected line in REPEAT")


def goto(ln):
    if len(ln) != 1:
        error("Usage: GOTO Label")
    gbl.inpath.goto(ln[0].upper())


def eof(ln):
        gbl.inpath.toEof()

#######################################
# File and I/O


def include(ln):
    """ Include a file. """

    global beginData

    if beginData:
        error("INCLUDE not permitted in Begin/End block")

    if len(ln) != 1:
        error("Use: Include FILE")

    fn = mma.MMA.paths.findIncFile(ln[0])

    if not fn:
        error("Could not find include file '%s'" % ln[0])

    parseFile(fn)


def usefile(ln):
    """ Include a library file. """

    global beginData

    if beginData:
        error("USE not permitted in Begin/End block")

    if len(ln) != 1:
        error("Use: Use FILE")

    fn = mma.MMA.paths.findLibFile(ln[0])

    if not fn:
        error("Unable to locate library file '%s'" % ln[0])

    """ USE saves current state, just like defining a groove.
        Here we use a magic number which can't be created with
        a defgroove ('cause it's an integer). Save, read, restore.
    """

    mma.MMA.grooves.stackGroove.push()
    parseFile(fn)
    mma.MMA.grooves.stackGroove.pop()


#######################################
# Misc

def rndseed(ln):
    """ Reseed the random number generator. """

    if not ln:
        random.seed()   # just resets, not predicable.

    elif len(ln) > 1:
        error("RNDSEED: requires 0 or 1 arguments")
    else:
        random.seed(stoi(ln[0]))  # predicable results.


def lnPrint(ln):
    """ Print stuff in a "print" command. """

    print(" ".join(ln))


def printActive(ln):
    """ Print a list of the active tracks. """

    print("Active tracks, groove: %s %s" % (mma.MMA.grooves.currentGroove, ' '.join(ln)))
    print("%15s  %2s   %s" % ("Track", "Ch", "Events"))
    for a in sorted(gbl.tnames.keys()):
        f = gbl.tnames[a]
        if f.sequence:
            if f.channel > 0:
                ch = f.channel
                ecount = 0
                for ev in gbl.mtrks[ch].miditrk:
                    ecount +=  len(gbl.mtrks[ch].miditrk[ev])
            else:
                ch = '-'
                ecount = ''
            print("%15s  %2s   %s" % (a, ch, ecount))
    print("\n")


###########################################################
###########################################################
## Track specific commands


#######################################
# Pattern/Groove

    
def trackDefPattern(name, ln):
    """ Define a pattern for a track.

    Use the type-name for all defines.... check the track
    names and if it has a '-' in it, we use only the
    part BEFORE the '-'. So DRUM-Snare becomes DRUM.
    """

    ln = ln[:]

    name = name.split('-')[0]

    trackAlloc(name, 1)

    if ln:
        pattern = ln.pop(0).upper()
    else:
        error("Define is expecting a pattern name")

    if pattern in ('z', 'Z', '-'):
        error("Pattern name '%s' is reserved" % pattern)

    if pattern.startswith('_'):
        error("Names with a leading underscore are reserved")

    if not ln:
        error("No pattern list given for '%s %s'" % (name, pattern))

    ln = ' '.join(ln)
    gbl.tnames[name].definePattern(pattern, ln)


def trackRiff(name, ln):
    """ Set a riff for a track. """

    gbl.tnames[name].setRiff(' '.join(ln))


def trackDupRiff(name, ln):
    """ Set a riff for a track. """

    if not ln:
        error("%s DupRiff: need at least one track to copy to.")

    gbl.tnames[name].dupRiff(ln)


def deleteTrks(ln):
    """ Delete a track and free the MIDI track. """

    if not len(ln):
        error("Use Delete Track [...]")

    for name in ln:
        name = name.upper()
        if name in gbl.tnames:
            tr = gbl.tnames[name]
        else:
            error("Track '%s' does not exist" % name)

        if tr.channel:
            tr.doMidiClear()
            tr.clearPending()

            if tr.riff:
                warning("%s has pending RIFF(s)" % name)
            gbl.midiAvail[tr.channel] -= 1

            # NOTE: Don't try deleting 'tr' since it's just a copy!!

            del gbl.tnames[name]

        if not name in gbl.deletedTracks:
            gbl.deletedTracks.append(name)

        if mma.MMA.debug.debug:
            dPrint("Track '%s' deleted" % name)

#######################################
# Volume


def trackRvolume(name, ln):
    """ Set random volume for specific track. """

    if not ln:
        error("Use: %s RVolume N [...]" % name)
    gbl.tnames[name].setRVolume(ln)


def trackSwell(name, ln):
    gbl.tnames[name].setSwell(ln)


def trackCresc(name, ln):
    gbl.tnames[name].setCresc(1, ln)


def trackDeCresc(name, ln):
    gbl.tnames[name].setCresc(-1, ln)


def trackVolume(name, ln):
    """ Set volume for specific track. """

    if not ln:
        error("Use: %s Volume DYN [...]" % name)

    gbl.tnames[name].setVolume(ln)


def trackChords(name, ln):
    """ Set a chord line for a specific track. """

    gbl.tnames[name].setChords(ln)


def trackAccent(name, ln):
    """ Set emphasis beats for track."""

    gbl.tnames[name].setAccent(ln)


#######################################
# Timing

def trackMallet(name, ln):
    """ Set repeating-mallet options for solo/melody track. """

    if not ln:
        error("Use: %s Mallet <Option=Value> [...]" % name)

    gbl.tnames[name].setMallet(ln)


def trackRduration(name, ln):
    """ Set random duration effect for specific track."""

    if not ln:
        error("Use: %s Rduration N [...]" % name)

    gbl.tnames[name].setRDuration(ln)


def trackRtime(name, ln):
    """ Set random timing for specific track. """

    if not ln:
        error("Use: %s RTime N [...]" % name)

    gbl.tnames[name].setRTime(ln)


def trackRskip(name, ln):
    """ Set random skip for specific track. """

    gbl.tnames[name].setRSkip(ln)


def trackArtic(name, ln):
    """ Set articulation. """

    if not ln:
        error("Use: %s Articulation N [...]" % name)

    gbl.tnames[name].setArtic(ln)


#######################################
# Chord stuff


def trackCompress(name, ln):
    """ Set (unset) compress for track. """

    if not ln:
        error("Use: %s Compress <value[s]>" % name)

    gbl.tnames[name].setCompress(ln)


def trackVoicing(name, ln):
    """ Set Voicing options. Only valid for chord tracks at this time."""

    if not ln:
        error("Use: %s Voicing <MODE=VALUE> [...]" % name)

    gbl.tnames[name].setVoicing(ln)


def trackDupRoot(name, ln):
    """ Set (unset) the root note duplication. Only applies to chord tracks. """

    if not ln:
        error("Use: %s DupRoot <value> ..." % name)

    gbl.tnames[name].setDupRoot(ln)


def trackChordLimit(name, ln):
    """ Set (unset) ChordLimit for track. """

    if len(ln) != 1:
        error("Use: %s ChordLimit <value>" % name)

    gbl.tnames[name].setChordLimit(ln[0])


def trackRange(name, ln):
    """ Set (unset) Range for track. Only effects arp and scale. """

    if not ln:
        error("Use: %s Range <value> ... " % name)

    gbl.tnames[name].setRange(ln)


def trackInvert(name, ln):
    """ Set invert for track."""

    if not ln:
        error("Use: %s Invert N [...]" % name)

    gbl.tnames[name].setInvert(ln)


def trackSpan(name, ln):
    """ Set midi note span for track. """

    if len(ln) != 2:
        error("Use: %s Start End" % name)

    start = stoi(ln[0], "Expecting integer for SPAN 1st arg")
    if start < 0 or start > 127:
        error("Start arg for Span must be 0..127, not %s" % start)

    end = stoi(ln[1], "Expecting integer for SPAN 2nd arg")
    if end < 0 or end > 127:
        error("End arg for Span must be 0..127, not %s" % end)

    if end <= start:
        error("End arg for Span must be greater than start")

    if end - start < 11:
        error("Span range must be at least 12")

    gbl.tnames[name].setSpan(start, end)


def trackOctave(name, ln):
    """ Set octave for specific track. """

    if not ln:
        error("Use: %s Octave N [...], (n=0..10)" % name)

    gbl.tnames[name].setOctave(ln)


def trackMOctave(name, ln):
    """ Set midi-based octave for specific track. """

    if not ln:
        error("Use: %s MOctave N [...], (n=01..9)" % name)

    gbl.tnames[name].setMOctave(ln)


def trackRpitch(name, ln):
    """ Set random pitch adjustment for specific track. """

    if not ln:
        error("Use: %s RPitch N [...]" % name)
    gbl.tnames[name].setRPitch(ln)

def trackStrum(name, ln):
    """ Set all specified track strum. """

    if not ln:
        error("Use: %s Strum N [...]" % name)

    gbl.tnames[name].setStrum(ln)

def trackStrumAdd(name, ln):
    """ Set all specified track strumAdd. """

    if not ln:
        error("Use: %s StrumAdd N [...]" % name)

    gbl.tnames[name].setStrumAdd(ln)

def trackSticky(name, ln):
    """ Sets a track as sticky. Ignored by groove commands. """

    if not ln:
        error("Use: %s Sticky On/Off" % name)

    gbl.tnames[name].setSticky(ln)

def trackHarmony(name, ln):
    """ Set harmony value. """

    if not ln:
        error("Use: %s Harmony N [...]" % name)

    mma.MMA.harmony.setHarmony(gbl.tnames[name], ln)
#    gbl.tnames[name].setHarmony(ln)


def trackHarmonyOnly(name, ln):
    """ Set harmony only for track. """

    if not ln:
        error("Use: %s HarmonyOnly N [...]" % name)

    mma.MMA.harmony.setHarmonyOnly(gbl.tnames[name], ln)
#      gbl.tnames[name].setHarmonyOnly(ln)


def trackHarmonyVolume(name, ln):
    """ Set harmony volume for track."""

    if not ln:
        error("Use: %s HarmonyVolume N [...]" % name)

    mma.MMA.harmony.setHarmonyVolume(gbl.tnames[name], ln)
#      gbl.tnames[name].setHarmonyVolume(ln)


#######################################
# Plectrum only stuff

def trackPlectrumTuning(name, ln):
    """ Define the number of strings and tuning for
        for an instrument that can be played with a plectrum.
    """

    if not ln:
        error("Use: %s Tuning string1 string2 string3 [stringN ...]" % name)

    g = gbl.tnames[name]

    try:
        g.setPlectrumTuning(ln)
    except AttributeError:
        warning("TUNING: not permitted in %s tracks. Arg '%s' ignored." %
                (g.vtype, ' '.join(ln)))


def trackPlectrumCapo(name, ln):
    """ Define the position of the capo
        (unlike a real guitar negative numbers are allowed)
        for an instrument that can be played with a plectrum.
    """

    if not ln or len(ln) != 1:
        error("Use: %s Capo N" % name)

    g = gbl.tnames[name]
    try:
        g.setPlectrumCapo(ln[0])
    except AttributeError:
        warning("CAPO: not permitted in %s tracks. Arg '%s' ignored." %
                (g.vtype, ' '.join(ln)))


def trackPlectrumFretNoise(name, ln):
    """ Define fret noise options.
    """

    g = gbl.tnames[name]

    try:
        g.setPlectrumFretNoise(ln)
    except AttributeError:
        warning("FRETNOISE: not permitted in %s tracks. Arg '%s' ignored." %
                (g.vtype, ' '.join(ln)))

def trackPlectrumShape(name, ln):
    """ Define chord shape for stringed instrument. """

    g = gbl.tnames[name]
    
    try:
        g.setPlectrumShape(ln)
    except AttributeError:
        warning("SHAPE: not permitted in %s tracks. Arg '%s' ignored." %
                (g.vtype, ' '.join(ln)))


#######################################
# MIDI setting


def trackChannel(name, ln):
    """ Set the midi channel for a track."""

    if not ln:
        error("Use: %s Channel" % name)

    gbl.tnames[name].setChannel(ln[0])


def trackChShare(name, ln):
    """ Set MIDI channel sharing."""

    if len(ln) != 1:
        error("Use: %s ChShare TrackName" % name)

    gbl.tnames[name].setChShare(ln[0])


def trackVoice(name, ln):
    """ Set voice for specific track. """

    if not ln:
        error("Use: %s Voice NN [...]" % name)

    gbl.tnames[name].setVoice(ln)


def trackOff(name, ln):
    """ Turn a track off """

    if ln:
        error("Use: %s OFF with no paramater" % name)

    gbl.tnames[name].setOff()


def trackOn(name, ln):
    """ Turn a track on """

    if ln:
        error("Use: %s ON with no paramater" % name)

    gbl.tnames[name].setOn()


def trackOrnament(name, ln):
    """ Set the ornamentation. Currently only for SCALE. """

    mma.MMA.ornament.setOrnament(gbl.tnames[name], ln)


def trackTone(name, ln):
    """ Set the tone (note). Only valid in drum tracks."""

    gbl.tnames[name].setTone(ln)


def trackForceOut(name, ln):
    """ Force output of voice settings. """

    if len(ln):
        error("Use %s ForceOut (no options)" % name)

    gbl.tnames[name].setForceOut()


#######################################
# Misc

def trackArpeggiate(name, ln):
    """ Set up the solo/melody arpeggiator. """

    if not ln:
        error("Use: %s Arpeggiate N" % name)

    g = gbl.tnames[name]
    try:
        g.setArp(ln)
    except AttributeError:
        warning("Arpeggiate: not permitted in %s tracks. Arg '%s' ignored." %
                (g.vtype, ' '.join(ln)))


def trackStretch(name, ln):
    """ Set the stretch value for solo/melody. """

    if not ln:
        error("Use: %s Stretch N" % name)

    g = gbl.tnames[name]
    try:
        g.setStretch(ln)
    except AttributeError:
        warning("Stretch: not permitted in %s tracks. Arg '%s' ignored." %
                (g.vtype, ' '.join(ln)))


def trackDelay(name, ln):
    """ Set up the solo/melody delay (echo). """

    if not ln:
        error("Use: %s Delay N" % name)

    gbl.tnames[name].setDelay(ln)


def trackDrumType(name, ln):
    """ Set a melody or solo track to be a drum solo track."""

    tr = gbl.tnames[name]
    if tr.vtype not in ('SOLO', 'MELODY'):
        error("Only Solo and Melody tracks can be to DrumType, not '%s'" % name)
    if ln:
        error("No parmeters permitted for DrumType command")

    tr.setDrumType()


def trackDirection(name, ln):
    """ Set scale/arp direction. """

    if not ln:
        error("Use: %s Direction OPT" % name)

    gbl.tnames[name].setDirection(ln)


def trackScaletype(name, ln):
    """ Set the scale type. """

    if not ln:
        error("Use: %s ScaleType OPT" % name)

    gbl.tnames[name].setScaletype(ln)


def trackUnify(name, ln):
    """ Set UNIFY for track."""

    if not len(ln):
        error("Use %s UNIFY 1 [...]" % name)

    gbl.tnames[name].setUnify(ln)


""" =================================================================

    Command jump tables. These need to be at the end of this module
    to avoid undefined name errors. The tables are only used in
    the parse() function.

    The first table is for the simple commands ... those which DO NOT
    have a leading trackname. The second table is for commands which
    require a leading track name.

    The alphabetic order is NOT needed, just convenient.

"""

simpleFuncs = {'ADJUSTVOLUME': mma.MMA.volume.adjvolume,
               'AFTER': mma.MMA.after.create,
               'ALLGROOVES': mma.MMA.grooves.allgrooves,
               'ALLTRACKS': allTracks,
               'AUTHOR': mma.MMA.docs.docAuthor,
               'AUTOSOLOTRACKS': mma.MMA.patSolo.setAutoSolo,
               'BEATADJUST': mma.MMA.tempo.beatAdjust,
               'CALL': mma.MMA.func.callFunction,
               'CHANNELPREF': mma.MMA.midifuncs.setChPref,
               'CHORDADJUST': mma.MMA.chords.chordAdjust,
               'CMDLINE': mma.MMA.options.cmdLine,
               'COMMENT': comment,
               'CRESC': mma.MMA.volume.setCresc,
               'CUT': mma.MMA.tempo.cut,
               'DEBUG': mma.MMA.debug.setDebug,
               'DEC': macros.vardec,
               'DECRESC': mma.MMA.volume.setDecresc,
               'DEFALIAS': mma.MMA.grooves.grooveAlias,
               'DEFCHORD': mma.MMA.chords.defChord,
               'DEFCALL': mma.MMA.func.defCall,
               'DEFGROOVE': mma.MMA.grooves.grooveDefine,
               'DELETE': deleteTrks,
               'DOC': mma.MMA.docs.docNote,
               'DOCVAR': mma.MMA.docs.docVars,
               'DRUMVOLTR': mma.MMA.translate.drumVolTable.set,
               'ELSE': ifelse,
               'ENDIF': ifend,
               'ENDMSET': endmset,
               'ENDREPEAT': repeatend,
               'EOF': eof,
               'FERMATA': mma.MMA.tempo.fermata,
               'GOTO': goto,
               'GROOVE': mma.MMA.grooves.groove,
               'GROOVECLEAR': mma.MMA.grooves.grooveClear,
               'IF': macros.varIF,
               'IFEND': ifend,
               'INC': macros.varinc,
               'INCLUDE': include,
               'KEYSIG': keySig.create,
               'LABEL': comment,
               'LYRIC': lyric.option,
               'MIDIDEF': mma.MMA.mdefine.mdefine,
               'MIDI': mma.MMA.midifuncs.rawMidi,
               'MIDICOPYRIGHT': mma.MMA.midifuncs.setMidiCopyright,
               'MIDICUE': mma.MMA.midifuncs.setMidiCue,
               'MIDIFILE': mma.MMA.midifuncs.setMidiFileType,
               'MIDIINC': mma.MMA.midiIn.midiinc,
               'MIDIVOLUME': mma.MMA.midifuncs.setMidiVolume,
               'MIDICRESC': mma.MMA.midifuncs.setMidiCresc,
               'MIDIDECRESC': mma.MMA.midifuncs.setMidiDecresc,
               'CHANNELINIT': mma.MMA.midifuncs.setChannelInit,
               'MIDIMARK': mma.MMA.midifuncs.midiMarker,
               'MIDISPLIT': mma.MMA.midi.setSplitChannels,
               'MIDITEXT': mma.MMA.midifuncs.setMidiText,
               'MIDITNAME': mma.MMA.midifuncs.setMidiName,
               'MMAEND': mma.MMA.paths.mmaend,
               'MMASTART': mma.MMA.paths.mmastart,
               'MSET': macros.msetvar,
               'MSETEND': endmset,
               'NEWSET': macros.newsetvar,
               'PATCH': mma.MMA.patch.patch,
               'PLUGIN': mma.MMA.regplug.plugin,
               'PRINT': lnPrint,
               'PRINTACTIVE': printActive,
               'PRINTCHORD': mma.MMA.chords.printChord,
               'REPEAT': repeat,
               'REPEATEND': repeatend,
               'REPEATENDING': repeatending,
               'RESTART': mma.MMA.sequence.restart,
               'RNDSEED': rndseed,
               'RNDSET': macros.rndvar,
               'SEQ': mma.MMA.sequence.seq,
               'SEQCLEAR': mma.MMA.sequence.seqClear,
               'SEQRND': mma.MMA.seqrnd.setSeqRnd,
               'SEQRNDWEIGHT': mma.MMA.seqrnd.setSeqRndWeight,
               'SEQSIZE': mma.MMA.sequence.seqsize,
               'SET': macros.setvar,
               'SETINCPATH': mma.MMA.paths.setIncPath,
               'SETLIBPATH': mma.MMA.paths.setLibPath,
               'SETMIDIPLAYER': mma.MMA.player.setMidiPlayer,
               'SETOUTPATH': mma.MMA.paths.setOutPath,
               'SETPLUGPATH': mma.MMA.paths.setPlugPath,
               'SETSYNCTONE': mma.MMA.sync.setSyncTone,
               'SHOWVARS': macros.showvars,
               'STACKVALUE': macros.stackValue,
               'SWELL': mma.MMA.volume.setSwell,
               'SWINGMODE': mma.MMA.swing.swingMode,
               'SYNCHRONIZE': mma.MMA.sync.synchronize,
               'TEMPO': mma.MMA.tempo.tempo,
               'TIME': mma.MMA.tempo.setTime,
               'TIMESIG': timeSig.setSig,
               'TONETR': mma.MMA.translate.dtable.set,
               'TRUNCATE': mma.MMA.truncate.setTruncate,
               'TWEAKS': mma.MMA.tweaks.setTweak,
               'UNSET': macros.unsetvar,
               'USE': usefile,
               'VARCLEAR': macros.clear,
               'VEXPAND': macros.vexpand,
               'VOICEVOLTR': mma.MMA.translate.voiceVolTable.set,
               'VOICETR': mma.MMA.translate.vtable.create,
               'VOLUME': mma.MMA.volume.setVolume,
               'TRANSPOSE': mma.MMA.keysig.transpose}

trackFuncs = {
    'ACCENT': trackAccent,
    'ARPEGGIATE': trackArpeggiate,
    'ARTICULATE': trackArtic,
    'CHANNEL': trackChannel,
    'CHORDS': trackChords,
    'DUPRIFF': trackDupRiff,
    'MIDIVOLUME': mma.MMA.midifuncs.trackMidiVolume,
    'MIDICRESC': mma.MMA.midifuncs.trackMidiCresc,
    'MIDIDECRESC': mma.MMA.midifuncs.trackMidiDecresc,
    'CHSHARE': trackChShare,
    'COMPRESS': trackCompress,
    'COPY': mma.MMA.grooves.trackCopy,
    'CRESC': trackCresc,
    'CUT': mma.MMA.tempo.trackCut,
    'DECRESC': trackDeCresc,
    'DELAY': trackDelay,
    'DIRECTION': trackDirection,
    'DRUMTYPE': trackDrumType,
    'DUPROOT': trackDupRoot,
    'FORCEOUT': trackForceOut,
    'FRETNOISE': trackPlectrumFretNoise,
    'GROOVE': mma.MMA.grooves.trackGroove,
    'HARMONY': trackHarmony,
    'HARMONYONLY': trackHarmonyOnly,
    'HARMONYVOLUME': trackHarmonyVolume,
    'INVERT': trackInvert,
    'LIMIT': trackChordLimit,
    'MALLET': trackMallet,
    'MIDICLEAR': mma.MMA.midifuncs.trackMidiClear,
    'MIDICUE': mma.MMA.midifuncs.trackMidiCue,
    'MIDIDEF': mma.MMA.mdefine.trackMdefine,
    'MIDIGLIS': mma.MMA.midifuncs.trackGlis,
    'MIDIPAN': mma.MMA.midifuncs.trackPan,
    'MIDISEQ': mma.MMA.midifuncs.trackMidiSeq,
    'MIDITEXT': mma.MMA.midifuncs.trackMidiText,
    'MIDITNAME': mma.MMA.midifuncs.trackMidiName,
    'MIDIVOICE': mma.MMA.midifuncs.trackMidiVoice,
    'MIDIWHEEL': mma.MMA.midifuncs.trackWheel,
    'MOCTAVE': trackMOctave,
    'OCTAVE': trackOctave,
    'OFF': trackOff,
    'ON': trackOn,
    "ORNAMENT": trackOrnament,
    'TUNING': trackPlectrumTuning,
    'CAPO': trackPlectrumCapo,
    'SHAPE': trackPlectrumShape,
    'RANGE': trackRange,
    'RDURATION': trackRduration,
    'RESTART': mma.MMA.sequence.trackRestart,
    'RIFF': trackRiff,
    'RSKIP': trackRskip,
    'RTIME': trackRtime,
    'RVOLUME': trackRvolume,
    'RPITCH':  mma.MMA.rpitch.setRPitch,
    'SCALETYPE': trackScaletype,
    'SEQCLEAR': mma.MMA.sequence.trackSeqClear,
    'SEQRND': mma.MMA.sequence.trackSeqRnd,
    'SEQUENCE': mma.MMA.sequence.trackSequence,
    'SEQRNDWEIGHT': mma.MMA.sequence.trackSeqRndWeight,
    'STRETCH': trackStretch,
    'STICKY': trackSticky,
    'SWELL': trackSwell,
    'TRIGGER': mma.MMA.trigger.setTrigger,
    'MIDINOTE': mma.MMA.midinote.parse,
    'NOTESPAN': trackSpan,
    'STRUM': trackStrum,
    'STRUMADD': trackStrumAdd,
    'TONE': trackTone,
    'UNIFY': trackUnify,
    'VOICE': trackVoice,
    'VOICING': trackVoicing,
    'VOLUME': trackVolume,
    'DEFINE': trackDefPattern}

dataFuncs = {}




# miditables.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

This module contains the constant names for the various
MIDI controllers.

Having only the constants in this separate file permits to
call this from other programs, mainly the mma doc creators.

"""

# Create a NONE value for "do nothing" midi voicing. This is 127.127.127
NONETONE = (127 << 16) + (127 < 8) + 127

#  Standard GM drum tone  names.

drumNames = {
    27: 'HighQ', 28: 'Slap', 29: 'ScratchPush',
    30: 'ScratchPull', 31: 'Sticks', 32: 'SquareClick',
    33: 'MetronomeClick', 34: 'MetronomeBell', 35: 'KickDrum2',
    36: 'KickDrum1', 37: 'SideKick', 38: 'SnareDrum1',
    39: 'HandClap', 40: 'SnareDrum2', 41: 'LowTom2',
    42: 'ClosedHiHat', 43: 'LowTom1', 44: 'PedalHiHat',
    45: 'MidTom2', 46: 'OpenHiHat', 47: 'MidTom1',
    48: 'HighTom2', 49: 'CrashCymbal1', 50: 'HighTom1',
    51: 'RideCymbal1', 52: 'ChineseCymbal', 53: 'RideBell',
    54: 'Tambourine', 55: 'SplashCymbal', 56: 'CowBell',
    57: 'CrashCymbal2', 58: 'VibraSlap', 59: 'RideCymbal2',
    60: 'HighBongo', 61: 'LowBongo', 62: 'MuteHighConga',
    63: 'OpenHighConga', 64: 'LowConga', 65: 'HighTimbale',
    66: 'LowTimbale', 67: 'HighAgogo', 68: 'LowAgogo',
    69: 'Cabasa', 70: 'Maracas', 71: 'ShortHiWhistle',
    72: 'LongLowWhistle', 73: 'ShortGuiro', 74: 'LongGuiro',
    75: 'Claves', 76: 'HighWoodBlock', 77: 'LowWoodBlock',
    78: 'MuteCuica', 79: 'OpenCuica', 80: 'MuteTriangle',
    81: 'OpenTriangle', 82: 'Shaker', 83: 'JingleBell',
    84: 'Castanets', 85: 'MuteSudro', 86: 'OpenSudro'}

drumInx = dict((v.upper(), k) for (k, v) in drumNames.items())

# Standard GM voice names.

voiceNames = {
    0: 'Piano1', 1: 'Piano2', 2: 'Piano3',
    3: 'Honky-TonkPiano', 4: 'RhodesPiano', 5: 'EPiano',
    6: 'HarpsiChord', 7: 'Clavinet', 8: 'Celesta',
    9: 'Glockenspiel', 10: 'MusicBox', 11: 'Vibraphone',
    12: 'Marimba', 13: 'Xylophone', 14: 'TubularBells',
    15: 'Santur', 16: 'Organ1', 17: 'Organ2',
    18: 'Organ3', 19: 'ChurchOrgan', 20: 'ReedOrgan',
    21: 'Accordion', 22: 'Harmonica', 23: 'Bandoneon',
    24: 'NylonGuitar', 25: 'SteelGuitar', 26: 'JazzGuitar',
    27: 'CleanGuitar', 28: 'MutedGuitar', 29: 'OverDriveGuitar',
    30: 'DistortionGuitar', 31: 'GuitarHarmonics', 32: 'AcousticBass',
    33: 'FingeredBass', 34: 'PickedBass', 35: 'FretlessBass',
    36: 'SlapBass1', 37: 'SlapBass2', 38: 'SynthBass1',
    39: 'SynthBass2', 40: 'Violin', 41: 'Viola',
    42: 'Cello', 43: 'ContraBass', 44: 'TremoloStrings',
    45: 'PizzicatoString', 46: 'OrchestralHarp', 47: 'Timpani',
    48: 'Strings', 49: 'SlowStrings', 50: 'SynthStrings1',
    51: 'SynthStrings2', 52: 'ChoirAahs', 53: 'VoiceOohs',
    54: 'SynthVox', 55: 'OrchestraHit', 56: 'Trumpet',
    57: 'Trombone', 58: 'Tuba', 59: 'MutedTrumpet',
    60: 'FrenchHorn', 61: 'BrassSection', 62: 'SynthBrass1',
    63: 'SynthBrass2', 64: 'SopranoSax', 65: 'AltoSax',
    66: 'TenorSax', 67: 'BaritoneSax', 68: 'Oboe',
    69: 'EnglishHorn', 70: 'Bassoon', 71: 'Clarinet',
    72: 'Piccolo', 73: 'Flute', 74: 'Recorder',
    75: 'PanFlute', 76: 'BottleBlow', 77: 'Shakuhachi',
    78: 'Whistle', 79: 'Ocarina', 80: 'SquareWave',
    81: 'SawWave', 82: 'SynCalliope', 83: 'ChifferLead',
    84: 'Charang', 85: 'SoloVoice', 86: '5thSawWave',
    87: 'Bass&Lead', 88: 'Fantasia', 89: 'WarmPad',
    90: 'PolySynth', 91: 'SpaceVoice', 92: 'BowedGlass',
    93: 'MetalPad', 94: 'HaloPad', 95: 'SweepPad',
    96: 'IceRain', 97: 'SoundTrack', 98: 'Crystal',
    99: 'Atmosphere', 100: 'Brightness', 101: 'Goblins',
    102: 'EchoDrops', 103: 'StarTheme', 104: 'Sitar',
    105: 'Banjo', 106: 'Shamisen', 107: 'Koto',
    108: 'Kalimba', 109: 'BagPipe', 110: 'Fiddle',
    111: 'Shanai', 112: 'TinkleBell', 113: 'AgogoBells',
    114: 'SteelDrums', 115: 'WoodBlock', 116: 'TaikoDrum',
    117: 'MelodicTom1', 118: 'SynthDrum', 119: 'ReverseCymbal',
    120: 'GuitarFretNoise', 121: 'BreathNoise', 122: 'SeaShore',
    123: 'BirdTweet', 124: 'TelephoneRing', 125: 'HelicopterBlade',
    126: 'Applause/Noise', 127: 'GunShot',
    NONETONE: 'None'}

voiceInx = dict((v.upper(), k) for (k, v) in voiceNames.items())
# fix an unfortunate spelling error.
voiceInx['DISTORTONGUITAR']=voiceInx['DISTORTIONGUITAR']

# Controller names. Tables are borrowed from:
#     http: //www.midi.org/about-midi/table3.shtml

#      0-31 Double Precise Controllers    MSB (14-bits, 16,384 values)
#      32-63  Double Precise Controllers  LSB (14-bits, 16,384 values)
#      64-119 Single Precise Controllers  (7-bits, 128 values)
#      120-127 Channel Mode Messages

ctrlNames = {
    0: 'Bank', 1: 'Modulation', 2: 'Breath',
    3: 'Ctrl3', 4: 'Foot', 5: 'Portamento',
    6: 'Data', 7: 'Volume', 8: 'Balance',
    9: 'Ctrl9', 10: 'Pan', 11: 'Expression',
    12: 'Effect1', 13: 'Effect2', 14: 'Ctrl14',
    15: 'Ctrl15', 16: 'General1', 17: 'General2',
    18: 'General3', 19: 'General4', 20: 'Ctrl20',
    21: 'Ctrl21', 22: 'Ctrl22', 23: 'Ctrl23',
    24: 'Ctrl24', 25: 'Ctrl25', 26: 'Ctrl26',
    27: 'Ctrl27', 28: 'Ctrl28', 29: 'Ctrl29',
    30: 'Ctrl30', 31: 'Ctrl31', 32: 'BankLSB',
    33: 'ModulationLSB', 34: 'BreathLSB', 35: 'Ctrl35',
    36: 'FootLSB', 37: 'PortamentoLSB', 38: 'DataLSB',
    39: 'VolumeLSB', 40: 'BalanceLSB', 41: 'Ctrl41',
    42: 'PanLSB', 43: 'ExpressionLSB', 44: 'Effect1LSB',
    45: 'Effect2LSB', 46: 'Ctrl46', 47: 'Ctrl47',
    48: 'General1LSB', 49: 'General2LSB', 50: 'General3LSB',
    51: 'General4LSB', 52: 'Ctrl52', 53: 'Ctrl53',
    54: 'Ctrl54', 55: 'Ctrl55', 56: 'Ctrl56',
    57: 'Ctrl57', 58: 'Ctrl58', 59: 'Ctrl59',
    60: 'Ctrl60', 61: 'Ctrl61', 62: 'Ctrl62',
    63: 'Ctrl63', 64: 'Sustain', 65: 'Portamento',
    66: 'Sostenuto', 67: 'SoftPedal', 68: 'Legato',
    69: 'Hold2', 70: 'Variation', 71: 'Resonance',
    72: 'ReleaseTime', 73: 'AttackTime', 74: 'Brightness',
    75: 'DecayTime', 76: 'VibratoRate', 77: 'VibratoDepth',
    78: 'VibratoDelay', 79: 'Ctrl79', 80: 'General5',
    81: 'General6', 82: 'General7', 83: 'General8',
    84: 'PortamentoCtrl', 85: 'Ctrl85', 86: 'Ctrl86',
    87: 'Ctrl87', 88: 'Ctrl88', 89: 'Ctrl89',
    90: 'Ctrl90', 91: 'Reverb', 92: 'Tremolo',
    93: 'Chorus', 94: 'Detune', 95: 'Phaser',
    96: 'DataInc', 97: 'DataDec', 98: 'NonRegLSB',
    99: 'NonRegMSB', 100: 'RegParLSB', 101: 'RegParMSB',
    102: 'Ctrl102', 103: 'Ctrl103', 104: 'Ctrl104',
    105: 'Ctrl105', 106: 'Ctrl106', 107: 'Ctrl107',
    108: 'Ctrl108', 109: 'Ctrl109', 110: 'Ctrl110',
    111: 'Ctrl111', 112: 'Ctrl112', 113: 'Ctrl113',
    114: 'Ctrl114', 115: 'Ctrl115', 116: 'Ctrl116',
    117: 'Ctrl117', 118: 'Ctrl118', 119: 'Ctrl119',
    120: 'AllSoundsOff', 121: 'ResetAll', 122: 'LocalCtrl',
    123: 'AllNotesOff', 124: 'OmniOff', 125: 'OmniOn',
    126: 'PolyOff', 127: 'PolyOn'}

ctrlInx = dict((v.upper(), k) for (k, v) in ctrlNames.items())



# midiIn.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

from . import gbl
import mma.MMA.file
import mma.MMA.debug
from   mma.MMA.common import *

from   mma.MMA.alloc import trackAlloc
from   mma.MMA.readmidi import MidiData

from   mma.MMA.midiM import packBytes

### FIXME ... some truly ugly code follows!!!

######################################################
## Main function, called from parser.

def midiinc(ln):
    """ Include a MIDI file into MMA generated files. """

    filename = ''
    doLyric = False
    doText = False
    channels = []
    transpose = None
    stripSilence = -1   # Important! Don't use True or 1 since positive # are option
    report = False
    istart = 0          # istart/end are in ticks
    iend = 0xffffff     # but are set in options in Beats
    verbose = False
    octAdjust = 0
    velAdjust = 100
    ignorePC = True
    stretch = None

    notopt, ln = opt2pair(ln)

    if notopt:
        error("MidiInc: Expecting cmd=opt pairs, not '%s'." % ' '.join(notopt))

    for cmd, opt in ln:
        cmd = cmd.upper()

        if cmd == 'FILE':
            filename = mma.MMA.file.fixfname(opt)

        elif cmd == 'VOLUME':
            velAdjust = stoi(opt)

        elif cmd == 'OCTAVE':
            octAdjust = stoi(opt)
            if octAdjust < -4 or octAdjust > 4:
                error("MidiInc: 'Octave' adjustment must be -4 to 4, not %s" % opt)
            octAdjust *= 12

        elif cmd == 'TRANSPOSE':
            transpose = stoi(opt)
            if transpose < -24 or transpose > 24:
                error("MidiInc: 'Transpose' must be -24 to 24, not %s" % opt)

        elif cmd == 'START':
            if opt[-1].upper() == 'M':   # measures
                istart = int(stof(opt[:-1]) * gbl.barLen)
            elif opt[-1].upper() == 'T':  # ticks
                istart = int(stof(opt[:-1]))
            else:  # must be digits, stof() catches errors
                istart = int((stof(opt)-1) * gbl.BperQ)

            if istart < 0:
                error("MidiInc: 'Start' must be > 0.")

        elif cmd == 'END':
            if opt[-1].upper() == 'M':
                iend = int((stof(opt[:-1])-1) * gbl.barLen)
            elif opt[-1].upper() == 'T':
                iend = int(stof(opt[:-1]))
            else:
                iend = int((stof(opt)-1) * gbl.BperQ)

            if iend < 0:
                error("MidiInc: 'End' must be > 0.")

        elif cmd == 'TEXT':
            opt = opt.upper()
            if opt in ("ON", 'TRUE', '1'):
                doText = True
            elif opt in ("OFF", 'FALSE', '0'):
                doText = False
            else:
                error("MidiInc: 'Text' expecting 'ON' or 'OFF'")

        elif cmd == 'LYRIC':
            opt = opt.upper()
            if opt in ("ON", 'TRUE', '1'):
                doLyric = True
            elif opt in ("OFF", 'FALSE', '0'):
                doLyric = False
            else:
                error("MidiInc: 'Lyric' expecting 'ON' or 'OFF'")

        elif cmd == "REPORT":
            opt = opt.upper()
            if opt in ("ON", 'TRUE', '1'):
                report = True
            elif opt in ("OFF", 'FALSE', '0'):
                report = False
            else:
                error("MidiInc: 'Report' expecting 'ON' or 'OFF'")

        elif cmd == "VERBOSE":
            opt = opt.upper()
            if opt in ("ON", 'TRUE','1'):
                verbose = True
            elif opt in ("OFF", 'FALSE', '0'):
                verbose = False
            else:
                error("MidiInc: 'Verbose' expecting 'ON' or 'OFF'")

        elif cmd == "STRIPSILENCE":
            opt = opt.upper()
            if opt in ("OFF", 'FALSE', '0'):
                stripSilence = False
            elif opt in ("TRUE", "ON"):  # this is the default
                stripSilence = -1  # Don't use TRUE, positive values are important
            else:
                stripSilence = stoi(opt, "MIdiInc StripSilence= expecting "
                                    "'value', 'On' or 'Off', not %s" % opt)

        elif cmd == "IGNOREPC":
            opt = opt.upper()
            if opt in ("TRUE", "ON", "1"):   # default
                ignorePC = True
            elif opt in ("FALSE", "OFF", "0"):  # use program change in imported
                ignorePC = False
            else:
                error("MIdiInc: 'IncludePC' expecting 'True' or 'False', not %s" % opt)

        elif cmd == "STRETCH":
            v = stof(opt)
            if v < 1 or v > 500:
                error("MidiInc: 'Stretch' range of 1 to 500, not %s." % opt)
            stretch = v/100.

        # If none of the above matched a CMD we assume that it is
        # a trackname. Keep this as the last test!

        else:
            trackAlloc(cmd, 0)
            if not cmd in gbl.tnames:
                error("MidiInc: %s is not a valid MMA track" % cmd)

            opt = opt.split(',')
            riffmode = 0
            printriff = 0
            ch = None
            for o in opt:
                o = o.upper()
                if o == 'RIFF':
                    riffmode = 1
                elif o == 'SEQUENCE':
                    riffmode = 2
                elif o == 'PRINT':
                    printriff = 1
                    if not riffmode:
                        riffmode = 1
                else:
                    if ch is not None:
                        error("MidiInc: Only one channel assignment per track.")
                    ch = stoi(o)

            if ch < 1 or ch > 16:
                error("MidiInc: MIDI channel for import must be 1..16, not %s" % ch)

            channels.append((cmd, ch-1, riffmode, printriff))

    # If transpose was NOT set, use the global transpose value
    # Note special riff value as well. Need to double adjust since
    # the riff import will do its own adjustment.
    # this needs to be done BEFORE reading the midi file

    if transpose is None:
        transpose = gbl.transpose
        riffTranspose = -gbl.transpose
    else:
        riffTranspose = 0
    octAdjust += transpose    # this takes care of octave and transpose

    
    mf = MidiData()
    mf.octaveAdjust = octAdjust
    mf.velocityAdjust = velAdjust
    mf.ignorePC = ignorePC

    try:
        mf.readFile(filename)
    except RuntimeError as e:
        error("MidiInc: %s" % e)

    if mf.beatDivision != gbl.BperQ:
        warning("MIDI file '%s' tick/beat of %s differs from MMA's "
                "%s. Will try to compensate" % 
                (filename, mf.beatDivision, gbl.BperQ))
        mf.adjustBeats( gbl.BperQ / float(mf.beatDivision))

    if report or verbose:  # try to be helpful
        print("MIDI File %s successfully read." % filename)
        print("Total Text events: %s" % len(mf.textEvents))
        print("Total Lyric events: %s" % len(mf.lyricEvents))
        print('\n')

        for ch in sorted(mf.events.keys()):
            if not mf.events[ch]:
                continue

            if verbose and not report:   # in verbose mode only list info for tracks we're using
                doit = 0
                for z in channels:
                    if z[1] == ch:
                        doit = 1
                        break
                
                if not doit:
                    continue

            fnote = fevent = 0xffffff
            ncount = 0
            for ev in mf.events[ch]:
                delta = ev[0]
                if delta < fevent:
                    fevent = delta
                if ev[1] >> 4 == 0x9:
                    if delta < fnote:
                        fnote = delta
                    if ev[3]:
                        ncount += 1
            msg = ["Channel %2s: First event %-8s" % (ch+1, fevent)]
            if ncount:
                msg.append("First Note %-8s Total Notes %-4s" % (fnote, ncount))
            print(' '.join(msg))

        if report:
            print("\nNo data generated!")
            sys.exit(0)
        
    if not channels:
        if doLyric or doText:
            warning("MidiInc: no import channels specified, "
                    "only text or lyrics imported")
        else:
            error("MidiInc: A channel to import and a destination "
                  "track must be specified")

    if (istart >= iend) or (istart < 0) or (iend < 0):
        error("MidiInc: Range invalid, start=%s, end=%s" % (istart, iend))

    if mma.MMA.debug.debug:
        dPrint("MidiInc: file=%s, Volume=%s, Octave=%s, Transpose=%s, Lyric=%s, " 
            "Text=%s, Range=%s..%s StripSilence=%s Verbose=%s" 
            % (filename, velAdjust, octAdjust, transpose, doLyric, doText,
               istart, iend, stripSilence, verbose))
        msg = []
        for t, ch, riffmode, riffprint in channels:
            o = ''
            if riffmode == 1:
                o = ',riff'
            elif riffmode == 2:
                o = ',sequence'
            elif printriff:
                o += ',print'
            msg.append("MidiInc: Channel %s-->%s%s" % (ch+1, t, o))
        dPrint(' '.join(msg))

    if stretch:
        if verbose:
            print("Applying stretch to all events. Deltas will be multiplied by %s" % stretch)

        for tr in mf.events:
            for e in mf.events[tr]:
                e[0] = int(e[0] * stretch)   # e[0] is the offset

        for e in mf.textEvents:
            e[0] = int(e[0] * stretch)

        for e in mf.lyricEvents:
            e[0] = int(e[0] * stretch)

    # Midi file parsed, add selected events to mma data

    if stripSilence == False:
        if verbose:
            print("Firstnote offset was %s. Being reset to start of file by StripSilence=Off." 
                % mf.firstNote)
        mf.firstNote = 0

    if verbose:
        print("First note offset: %s" % mf.firstNote)

    if doText:
        inst = 0
        disc = 0
        if verbose:
            print("Scanning %s textevents." % len(mf.textEvents))
        for tm, tx in mf.textEvents:
            delta = tm-mf.firstNote
            if delta >= istart and delta <= iend:
                gbl.mtrks[0].addText(gbl.tickOffset + delta, tx)
                inst += 1
            else:
                disc += 1

        if mma.MMA.debug.debug:
            dPrint("MidiInc text events: %s inserted, %s out of range." % (inst, disc))

    if doLyric:
        inst = 0
        disc = 0
        if verbose:
            print("Scanning %s LyricEvents." % len(mf.lyricEvents))
        for tm, tx in mf.lyricEvents:
            delta = tm-mf.firstNote
            if delta >= istart and delta <= iend:
                gbl.mtrks[0].addLyric(gbl.tickOffset + delta, tx)
                inst += 1
            else:
                disc += 1
        if mma.MMA.debug.debug:
            dPrint("MidiInc lyric events: %s inserted, %s out of range." % (inst, disc))

    for n, c, riffmode, printriff in channels:
        if not len(mf.events[c]):
            warning("No data to assign from imported channel %s to track %s" % (c+1, n))

    inst = 0
    disc = 0

    for tr, ch, riffmode, printriff in channels:
        onNotes = []
        if gbl.tnames[tr].disable:   # skip if disabled track

            if verbose:
                print("Skipping import of channel %s since track %s is disabled." 
                    % (ch, tr))
            continue

        t = gbl.tnames[tr]
        if not t.channel:
            t.setChannel()

        if riffmode:
            riff = []
            if t.vtype not in ('MELODY', 'SOLO'):
                error("MidiInc: Riff only works on Melody/Solo tracks, not '%s'." % t.name)

        t.clearPending()
        if t.voice[0] != t.ssvoice:
            gbl.mtrks[t.channel].addProgChange(gbl.tickOffset, t.voice[0], t.ssvoice)

        channel = t.channel
        track = gbl.mtrks[channel]

        if verbose:
                print("Parsing imported file. Channel=%s Track=%s MIDI Channel=%s" 
                    % (ch, tr, channel))
                if len(mf.events[ch]):
                    print(" Total events: %s; Event range: %s %s; Start/End Range: %s %s" 
                    % (len(mf.events[ch]), mf.events[ch][0][0], 
                       mf.events[ch][-1][0], istart, iend))
                else:
                    print("No events in Channel %s" % ch)

        
        # If we're processing midi voice changes (ignorePC=off) and there
        # are events BEFORE the first note, we need to insert
        # them before the notes. We put them all at the current midi offset.
        if ignorePC==0:
            for ev in mf.events[ch]:
                if ev[0] > mf.firstNote:
                    break
                if ev[1] >> 4 == 0xc:
                    track.addToTrack(gbl.tickOffset,
                                     packBytes(ev[1] | channel-1, *ev[2:]))
                    inst += 1
                    disc -= 1

        for ev in mf.events[ch]:
            delta = ev[0]-mf.firstNote
            
            if delta >= istart and delta <= iend:
                if riffmode:
                    offset = delta-istart
                    x = ev[1] >> 4
                    if x != 0x09 and x != 0x08:  # skip non note events
                        continue
                    pitch = ev[2]
                    velocity = ev[3]
                    if x == 0x8:
                        velocity = 0
                    riff.append([offset, pitch, velocity])

                else:
                    offset = gbl.tickOffset + (delta-istart)
                    # add note on/off, key pressure, etc.
                    track.addToTrack(offset, packBytes(ev[1] | channel-1, *ev[2:]))

                    # Track on/off events to avoid stuck midi notes.
                    x = ev[1] >> 4
                    if x == 0x09 and ev[3] and ev[2] not in onNotes:
                        onNotes.append(ev[2])  # append this pitch to onNotes
                    if (x == 0x09 and not ev[3] or x == 0x08) and (ev[2] in onNotes):
                        onNotes.remove(ev[2])  # remove this as being ON
                        
                inst += 1
            else:
                disc += 1

        if onNotes:
            for x in onNotes:
                track.addToTrack(offset, packBytes(0x90 | channel-1, [x,0]))
            warning("MidiINC: Stuck MIDI note(s) '%s' turned off in %s." % 
                    (', '.join([str(x) for x in onNotes]), tr))

        if riffmode:
            evlist = createRiff(riff, tr, riffTranspose)

            if riffmode == 2:
                txt = []
            for a in sorted(evlist):
                if printriff and riffmode == 1:
                    print("%s Riff %s" % (tr, evlist[a]))
                elif riffmode == 2:   # sequence mode, create sequence line and push into input
                    txt.append("{%s}" % evlist[a])
                else:   # riffmode==1, printriff=0 - just add to the riff stack
                    gbl.tnames[tr].setRiff(evlist[a])

            if riffmode == 2 and txt:
                if printriff:
                    print("%s Sequence %s" % (tr, ' '.join(txt)))
                else:
                    mma.MMA.sequence.trackSequence(tr, txt)

    if mma.MMA.debug.debug:
            dPrint("MidiInc events: %s inserted, %s out of range." % (inst, disc))


def createRiff(riff, tname, riffTranspose):

    # convert list of ON values to durations. We need to
    # look at each event and, if an on-event, search forward
    # for an off. Subtract 2 times and save in new list.

    if gbl.tnames[tname].riff:
        error("MidiInc: Data already pending for %s." % tname)

    missed = 0
    events = []
    riff.sort()
    for i in range(len(riff)):
        duration = None
        offset, pitch, velocity = riff[i]
        if velocity:
            for t in range(i, len(riff)):
                off1, pitch1, vel1 = riff[t]
                if not vel1 and pitch1 == pitch:
                    duration = off1 - offset
                    break
            if duration:
                if riffTranspose:
                    pitch += riffTranspose
                    while pitch > 127:
                        pitch -= 12
                    while pitch < 0:
                        pitch += 12
                events.append([offset, duration, pitch, velocity])

            else:
                missed += 1

    if missed:
        warning("MidiInc Riff: conversion missed %s notes in track %s" % (missed, tname))

    # We have a list of events: [offset, duration, pitch, velocity]...
    # create yet another list with the events put into bars. Easier
    # this time to use a dict

    tickBar = gbl.barLen
    bars = {}
    for offset, duration, pitch, velocity in events:
        b = (offset // tickBar)
        if not b in bars:
            bars[b] = ''
        if int(offset % tickBar) + duration > gbl.barLen:
            eol = "~;"
        else:
            eol = ";"
        bars[b] += "<Offset=%s> %st %s/%s %s" % \
            (int(offset % tickBar), duration, pitch, velocity, eol)

    # Ensure that all bars in the riff pushback data have something.
    # Otherwise the MIDI gets out of sync with the chords. This happens
    # when there are rest bars in the imported data. We just step though
    # and create mma rest bars for the missing ones.

    if not len(bars):
        bars = {0: "4r;"}
    for b in range(0, sorted(bars.keys())[-1], 1):
        if not b in bars:
            bars[b] = "4r;"

        while bars[b].count('~') > 1:
            bars[b] = bars[b].replace('~', '', 1)

    return bars



# patAria.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>

"""

import random

import mma.MMA.notelen
import mma.MMA.harmony
from mma.MMA.keysig import keySig
import mma.MMA.chords
import mma.MMA.debug
from . import gbl
from mma.MMA.common import *
from mma.MMA.pat import PC, Pgroup


class Aria(PC):
    """ Pattern class for an aria (auto-melody) track. """

    vtype = 'ARIA'
    notes = []
    selectDir = [1]
    noteptr = 0
    dirptr = 0
    lastChord = None
    deplete = [0]

    def setSeqSize(self):
        """ Expand existing pattern list. """

        self.deplete = seqBump(self.deplete)
        PC.setSeqSize(self)

    def restoreGroove(self, gname):
        """ Grooves are not saved/restored for aria tracks. But, seqsize is honored! """
        self.setSeqSize()

    def saveGroove(self, gname):
        """ No save done for grooves. """
        pass

    def getPgroup(self, ev):
        """ Get group for aria pattern.

            Fields - start, length, velocity

        """

        if len(ev) != 3:
            error("%s: There must be n groups of 3 in a pattern definition, "
                  "not <%s>" % (self.name, ' '.join(ev)))

        a = Pgroup()

        a.offset = self.setBarOffset(ev[0])
        a.duration = mma.MMA.notelen.getNoteLen(ev[1])
        a.vol = stoi(ev[2], "Note volume in Aria definition not int")

        return a

    def setScaletype(self, ln):
        """ Set scale type. """

        ln = lnExpand(ln, "%s ScaleType" % self.name)
        tmp = []
        dlpt = []

        for n in ln:
            n = n.upper()
            if n.endswith('-'):
                dlpt.append(1)
                n = n[:-1]
            else:
                dlpt.append(0)

            if not n in ('CHROMATIC', 'SCALE', 'AUTO', 'CHORD', 'KEY'):
                error("%s ScaleType: Only Chromatic, Scale (Auto) Chord "
                      "and Key are valid." % self.name)
            tmp.append(n)

        self.scaleType = seqBump(tmp)
        self.deplete = seqBump(dlpt)
        self.restart()

        if mma.MMA.debug.debug:
            msg = ["Set %s ScaleType:" % self.name]
            for a in self.scaleType:
                msg.append(a)
            dPrint(' '.join(msg))

    def setDirection(self, ln):
        """ Set direction for melody creation.

            This function replaces the pattern function of the same name ...
            the command name is shared, the function is different. Note we
            need to use a different storage name as well since
            self.direction is managed in the PC class.
        """

        if not len(ln):
            error("%s Direction: There must be at least one value." % self.name)

        self.selectDir = []
        for a in ln:
            if set(a.upper()) == set('R'):    # is direction 'r', 'rr', 'rrr', etc.
                if len(a) > 4:
                    error("%s Direction: too much randomness"
                          "(Maximum of 4 r's, got %d)." % (self.name, len(a)))
                self.selectDir.append(a.upper())
            else:   # not random, has to be an integer -4 ... 4
                a = stoi(a, "Expecting integer value or 'r', 'rr', 'rrr' or 'rrrr'.")
                if a < -4 or a > 4:
                    error("%s Direction: args must be 'r' or -4 to 4, not '%s'" % (self.name, a))
                self.selectDir.append(a)

        self.restart()

        if mma.MMA.debug.debug:
            msg = ["Set %s Direction:" % self.name]
            for a in self.selectDir:
                msg.append(str(a))
            dPrint(' '.join(msg))

    def restart(self):
        self.ssvoice = -1
        self.notes = []

    def trackBar(self, pattern, ctable):
        """ Do the aria bar.

        Called from self.bar()

        """

        sc = self.seq
        unify = self.unify[sc]

        for p in pattern:
            ct = self.getChordInPos(p.offset, ctable)

            if ct.ariaZ:
                continue

            thisChord = ct.chord.tonic + ct.chord.chordType
            stype = self.scaleType[sc]
            chrange = self.chordRange[sc]

            # Generate notelist if nesc. Note that in the keysig, scale and
            # range funcs restart() is called ... self.notes is reset.

            if stype == 'CHORD' and (not self.notes or self.lastChord != thisChord):
                notelist = ct.chord.noteList
                self.notes = []

            elif stype == 'CHROMATIC' and (not self.notes or self.lastChord != thisChord):
                notelist = [ct.chord.rootNote + x for x in range(0, 12)]
                self.notes = []

            elif stype == 'KEY' and not self.notes:
                k = keySig.getKeysig()
                ch, t = k.split()
                if t.lower() == 'minor':
                    ch += "m"
                notelist = list(mma.MMA.chords.ChordNotes(ch).scaleList)

            elif (stype == 'SCALE' or stype == 'AUTO') and \
                    (not self.notes or self.lastChord != thisChord):
                notelist = list(ct.chord.scaleList)
                self.notes = []

            self.lastChord = thisChord

            # we have the base list of notes (scale, chord, etc) and
            # now we make it the right length & octave.
            if not self.notes:
                o = 0
                while chrange >= 1:
                    for a in notelist:
                        self.notes.append(a + o)
                    o += 12
                    chrange -= 1

                if chrange > 0 and chrange < 1:  # for fractional scale lengths
                    chrange = int(len(notelist) * chrange)
                    if chrange < 2:   # important, must be at least 2 notes in a scale
                        chrange = 2
                    for a in notelist[:chrange]:
                        self.notes.append(a + o)

            # grab a note from the list

            if self.dirptr >= len(self.selectDir):
                self.dirptr = 0

            # the direction ptr is either an int(-4..4) or a string of 'r', 'rr, etc.

            a = self.selectDir[self.dirptr]

            if isinstance(a, int):
                self.noteptr += a
            else:
                a = random.choice(range(-len(a), len(a) + 1))
                self.noteptr += a

            if self.noteptr >= len(self.notes):

                if a > 0:
                    self.noteptr = 0
                else:
                    self.noteptr = len(self.notes) - 1
            elif self.noteptr < 0:
                if a < 0:
                    self.noteptr = len(self.notes) - 1
                else:
                    self.noteptr = 0

            note = self.notes[self.noteptr]

            # delete note just selected if that's the mode
            if self.deplete[sc]:
                self.notes.remove(note)
            self.dirptr += 1

            # output

            if not self.harmonyOnly[sc]:
                notelist = [(note, p.vol)]
            else:
                notelist = []

            if self.harmony[sc]:
                h = mma.MMA.harmony.harmonize(self.harmony[sc], note, ct.chord.noteList)
                harmlist = list(zip(h, [p.vol * self.harmonyVolume[sc]] * len(h)))
            else:
                harmlist = []

            self.sendChord(notelist + harmlist, p.duration, p.offset)



# sequence.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>


"""

from . import gbl
from mma.MMA.common import *
import mma.MMA.debug
from mma.MMA.notelen import noteLenTable

from mma.MMA.alloc import trackAlloc
from mma.MMA.macro import macros
from mma.MMA.pat import pats

import mma.MMA.seqrnd

def trackSequence(name, ln):
    """ Define a sequence for a track.

    The format for a sequence:
    TrackName Seq1 [Seq2 ... ]

    Note, that SeqX can be a predefined seq or { seqdef }
    The {} is dynamically interpreted into a def.
    """

    if not ln:
        error("Use: %s Sequence NAME [...]" % name)

    ln = ' '.join(ln)
    self = gbl.tnames[name]  # this is the pattern class

    if self.vtype == "SOLO":
            warning("Sequences for SOLO tracks are not saved in Grooves.")

    """ Before we do extraction of {} stuff make sure we have matching {}s.
        Count the number of { and } and if they don't match read more lines and
        append. If we get to the EOF then we're screwed and we error out. Only trick
        is to make sure we do macro expansion! This code lets one have long
        sequence lines without bothering with '\' continuations.
    """

    oLine = gbl.lineno   # in case we error out, report start line
    while ln.count('{') != ln.count('}'):
        l = gbl.inpath.read()
        if l is None:   # reached eof, error
            gbl.lineno = oLine
            error("%s Sequence {}s do not match" % name)

        l = ' '.join(macros.expand(l))

        if l[-1] != '}' and l[-1] != ';':
            error("%s: Expecting multiple sequence lines to end in ';'" % name)

        ln += ' ' + l

    """ Extract out any {} definitions and assign them to new
        define variables (__1, __99, etc) and melt them
        back into the string.
    """

    ids = 1

    while 1:
        sp = ln.find("{")

        if sp < 0:
            break

        ln, s = pextract(ln, "{", "}", onlyone=True)
        if not s:
            error("Did not find matching '}' for '{'")

        pn = "_%s" % ids
        ids += 1

        trk = name.split('-')[0]
        trackAlloc(trk, 1)

        """ We need to mung the plectrum classes. Problem is that we define all
            patterns in the base class (plectrum-banjo is created in PLECTRUM)
            which is fine, but the def depends on the number of strings in the
            instrument (set by the tuning option). So, we save the tuning for
            the base class, copy the real tuning, and restore it.

            NOTE: at this point the base and current tracks have been initialized.
        """

        if trk == 'PLECTRUM' and name != trk:
            z = gbl.tnames[trk]._tuning[:]
            gbl.tnames[trk]._tuning = gbl.tnames[name]._tuning
        else:
            z = None

        gbl.tnames[trk].definePattern(pn, s[0])  # 'trk' is a base class!
        if z:
            gbl.tnames[trk]._tuning = z

        ln = ln[:sp] + ' ' + pn + ' ' + ln[sp:]

    ln = ln.split()

    """ We now have a sequence we can save for the track. All the {} defs have
        been converted to special defines (_1, _2, etc.).

        First we expand ln to the proper length. lnExpand() also
        duplicates '/' to the previous pattern.

        Then we step though ln:

          - convert 'z', 'Z' and '-' to empty patterns.

          - duplicate the existing pattern for '*'

          - copy the defined pattern for everything else.
            There's a bit of Python reference trickery here.
            Eg, if we have the line:

              Bass Sequence B1 B2

            the sequence is set with pointers to the existing
            patterns defined for B1 and B2. Now, if we later change
            the definitions for B1 or B2, the stored pointer DOESN'T
            change. So, changing pattern definitions has NO EFFECT.

    """

    ln = lnExpand(ln, '%s Sequence' % self.name)
    tmp = [None] * len(ln)

    for i, n in enumerate(ln):
        n = n.upper()

        if n in ('Z', '-'):
            tmp[i] = None

        elif n == '*':
            tmp[i] = self.sequence[i]

        else:
            p = (self.vtype, n)
            if not p in pats:
                error("Track %s does not have pattern '%s'" % p)
            tmp[i] = pats[p]

    self.sequence = seqBump(tmp)

    if mma.MMA.debug.seqshow:
        msg = ["%s sequence set:" % self.name]
        for a in ln:
            if a in "Zz-":
                msg.append("-")
            else:
                msg.append(a)
        dPrint(' '.join(msg))


def seqsize(ln):
    """ Set the length of sequences. """

    if len(ln) != 1:
        error("Usage 'SeqSize N'")

    n = stoi(ln[0], "Argument for SeqSize must be integer")

    if n < 1:
        error("SeqSize: sequence size must be 1 or greater, not '%s'." % n)

    # Setting the sequence size always resets the seq point

    gbl.seqCount = 0

    """ Now set the sequence size for each track. The class call
        will expand/contract existing patterns to match the new
        size.
    """

    if n != gbl.seqSize:
        gbl.seqSize = n
        for a in gbl.tnames.values():
            a.setSeqSize()

        mma.MMA.seqrnd.seqRndWeight = seqBump(mma.MMA.seqrnd.seqRndWeight)

    if mma.MMA.debug.debug:
        dPrint("Set SeqSize to %s" % n)


def seq(ln):
    """ Set the sequence point. """

    if len(ln) == 0:
        s = 0
    elif len(ln) == 1:
        s = stoi(ln[0], "Expecting integer value after SEQ")
    else:
        error("Use: SEQ or SEQ NN to reset seq point")

    if s > gbl.seqSize:
        error("Sequence size is '%d', you can't set to '%d'" %
              (gbl.seqSize, s))

    if s == 0:
        s = 1

    if s < 0:
        error("Seq parm must be greater than 0, not %s", s)

    gbl.seqCount = s - 1

    if mma.MMA.seqrnd.seqRnd[0] == 1:
        warning("SeqRnd has been disabled by a Seq command")
        mma.MMA.seqrnd.seqRnd = [0]


def seqClear(ln):
    """ Clear all sequences (except SOLO/ARIA and STICKY tracks). """

    if ln:
        error("Use: 'SeqClear' with no args")

    for n in gbl.tnames.values():
        if n.vtype in ('SOLO', 'ARIA') or n.sticky:
            continue
        n.clearSequence()

    mma.MMA.volume.futureVol = []

    mma.MMA.seqrnd.setSeqRndWeight(['1'])


def restart(ln):
    """ Restart all tracks to almost-default conditions. """

    if ln:
        error("Use: 'Restart' with no args")

    for n in gbl.tnames.values():
        n.restart()

#####################################################
## Misc track sequence commands. Called from parser.

def trackSeqClear(name, ln):
    """ Clear sequence for specified tracks.

    Note: "Drum SeqClear" clears all Drum tracks,
          "Drum-3 SeqClear" clears track Drum-3.
    """

    if ln:
        error("No args permitted. Use %s SEQCLEAR" % name)

    for n in gbl.tnames:
        if n.find(name) == 0:
            if mma.MMA.debug.debug:
                dPrint("SeqClear: Track %s cleared." % n)
            gbl.tnames[n].clearSequence()


def trackSeqRnd(name, ln):
    """ Set random order for specified track. """

    if len(ln) != 1:
        error("Use: %s SeqRnd [On, Off]" % name)

    self = gbl.tnames[name]
    arg = ln[0].upper()

    if arg in ("TRUE", "ON", "1"):
        self.seqRnd = 1

    elif arg in ("FALSE", "OFF", "0"):
        self.seqRnd = 0

    else:
        error("SeqRnd: '%s' is not a valid option" % arg)

    if mma.MMA.debug.debug:
        if self.seqRnd:
            a = "On"
        else:
            a = "Off"
        dPrint("%s SeqRnd: %s" % (self.name, a))


def trackSeqRndWeight(name, ln):
    """ Set rnd weight for track. """

    if not ln:
        error("Use: %s SeqRndWeight <weight factors>" % name)

    self = gbl.tnames[name]


    ln = lnExpand(ln, "%s SeqRndWeight" % self.name)
    tmp = []

    for n in ln:
        n = stoi(n)
        if n < 0:
            error("SeqRndWeight: Values must be 0 or greater")
        tmp.append(n)

    self.seqRndWeight = seqBump(tmp)

    if mma.MMA.debug.debug:
        dPrint("Set %s SeqRndWeight: %s" % 
              (self.name, ' '.join([str(a) for a in self.seqRndWeight])))


def trackRestart(name, ln):
    """ Restart track to almost-default condidions. """

    if ln:
        error("Use: '%s Resart' with no args", name)

    gbl.tnames[name].restart()

